# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation
#               and/or initial documentation
#        :author: Sobieski, Agte, and Sandusky
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#        :author: Damien Guenot
#        :author: Francois Gallard
# From NASA/TM-1998-208715
# Bi-Level Integrated System Synthesis (BLISS)
# Sobieski, Agte, and Sandusky
"""
SSBJ Structure computations
***************************
"""
from __future__ import division, unicode_literals

import logging
from math import pi

from numpy import append, array, ones, zeros

LOGGER = logging.getLogger(__name__)
DEG_TO_RAD = pi / 180.0


class SobieskiStructure(object):
    """Class defining structural analysis for Sobieski problem and related method to the
    sructural problem such as disciplines computation, constraints, reference
    optimum."""

    DTYPE_COMPLEX = "complex128"
    DTYPE_DOUBLE = "float64"

    STRESS_LIMIT = 1.09
    TWIST_UPPER_LIMIT = 1.04
    TWIST_LOWER_LIMIT = 0.8

    def __init__(self, sobieski_base):
        """Constructor.

        :param sobieski_base: Sobieski problem
        :type sobieski_base: SobieskiBase
        """
        self.base = sobieski_base
        self.constants = self.base.default_constants()
        (
            self.x_initial,
            self.tc_initial,
            self.half_span_initial,
            self.aero_center_initial,
            self.cf_initial,
            self.mach_initial,
            self.h_initial,
            self.throttle_initial,
            self.lift_initial,
            self.twist_initial,
            self.esf_initial,
        ) = self.base.get_initial_values()
        self.dtype = self.base.dtype
        self.math = self.base.math

    def __set_coeff_twist(self, x_1, y_21, half_span, aero_center):
        """Prepare settings of polynomial function for wing twist computation.

        :param x_1: local design variables
        :type x_1: ndarray
        :param y_21: coupling variables from aerodynamics
        :type y_21: ndarray
        :param half_span: half-span of wing
        :type half_span: ndarray
        :param aero_center: aerodynamic center
        :type aero_center: ndarray
        :returns:  s_initial, s_new, flag, bound
        :rtype: ndarray
        """
        s_initial = array(
            [
                self.x_initial,
                self.half_span_initial,
                self.aero_center_initial,
                self.lift_initial,
            ]
        )
        s_new = array([x_1[1], half_span, aero_center, y_21[0]], dtype=self.dtype)
        flag = array([2, 4, 4, 3], dtype=self.dtype)
        bound = array([0.25, 0.25, 0.25, 0.25], dtype=self.dtype)
        return s_initial, s_new, flag, bound

    def __set_coeff_secthick(self, x_1):
        """Prepare settings of polynomial function for sectional thickness.

        :param x_1: local design variables
        :type x_1: ndarray
        :returns:  s_initial, s_new, flag, bound
        :rtype: ndarray
        """
        s_initial = array([self.x_initial], dtype=self.dtype)
        s_new = array([x_1[1]], dtype=self.dtype)
        flag = array([1], dtype=self.dtype)
        bound = array([0.008], dtype=self.dtype)
        return s_initial, s_new, flag, bound

    def __set_coeff_stress(self, x_1, x_shared, y_21, half_span, aero_center):
        """Prepare settings of polynomial function for stress.

        :param x_1: local design variables
        :type x_1: ndarray
        :param x_shared: system design variables
        :type x_shared: ndarray
        :param y_21: coupling variables from aerodynamics
        :type y_21: ndarray
        :param half_span: half-span of wing
        :type half_span: ndarray
        :param aero_center: aerodynamic center
        :type aero_center: ndarray
        :returns:  s_initial, s_new, flag, bound
        :rtype: ndarray
        """
        s_initial = array(
            [
                self.tc_initial,
                self.lift_initial,
                self.x_initial,
                self.half_span_initial,
                self.aero_center_initial,
            ],
            dtype=self.dtype,
        )
        s_new = array(
            [x_shared[0], y_21[0], x_1[1], half_span, aero_center], dtype=self.dtype
        )
        flag = array([4, 1, 4, 1, 1], dtype=self.dtype)
        return s_initial, s_new, flag

    def __compute_dadimcenter_dcenter(self, x_1):
        """Compute derivative of adim aero center wrt aero center.

        :param x_1: local design variables
        :type x_1: ndarray
        :returns:  derivative of adim aero center wrt aero center
        :rtype: ndarray
        """
        aero_center = self.base.compute_aero_center(x_1)
        return self.base.derive_normalize_s(self.aero_center_initial, aero_center)

    def __compute_dcenter_dlambda(self, x_1):
        """Compute derivative of aero center wrt wing taper ratio.

        :param x_1: local design variables
        :type x_1: ndarray
        :returns:  dR_dlambda
        :rtype: ndarray
        """
        dadimcenter_dcenter = self.__compute_dadimcenter_dcenter(x_1)
        return dadimcenter_dcenter * 1.0 / (3.0 * (1.0 + x_1[0]) ** 2)

    def compute_wing_weight(self, x_shared, x_1, y_21, linearize=False):
        """Compute wing weight.

        :param x_shared: system design variables
        :type x_shared: ndarray
        :param y_21: coupling variables from aerodynamics
        :type y_21: ndarray
        :param x_1: local design variables
        :type x_1: ndarray
        :param linearize: Default value = False)
        :returns: wing weight, wing_weight_coeff, value of poly. function
        :rtype: ndarray
        """
        s_initial, s_new, flag, bound = self.__set_coeff_secthick(x_1)
        if linearize:
            f_o, a_i, a_ij, s_shifted = self.base.derive_poly_approx(
                s_initial, s_new, flag, bound
            )
        else:
            f_o = self.base.poly_approx(s_initial, s_new, flag, bound)
        wing_weight_coeff = (
            0.0051
            * ((y_21[0] * self.constants[2]) ** 0.557)
            * (x_shared[5] ** 0.649)
            * (x_shared[3] ** 0.5)
            * (x_shared[0] ** -0.4)
            * ((1 + x_1[0]) ** 0.1)
            * ((self.math.cos(x_shared[4] * DEG_TO_RAD)) ** -1.0)
            * ((0.1875 * x_shared[5]) ** 0.1)
        )
        wing_weight = wing_weight_coeff * f_o
        if linearize:
            return wing_weight, wing_weight_coeff, f_o, a_i, a_ij, s_shifted
        return wing_weight

    def compute_fuelwing_weight(self, x_shared):
        """Compute fuel wing weight.

        :param x_shared: system design variables
        :type x_shared: ndarray
        :returns: fuel wing weight
        :rtype: ndarray
        """
        thickness = self.base.compute_thickness(x_shared)  # wing thickness
        return (5.0 * x_shared[5] / 18.0) * (2.0 / 3.0 * thickness) * 42.5

    def compute_dfuelwing_dtoverc(self, x_shared):
        """Compute fuel wing weight.

        :param x_shared: system design variables
        :type x_shared: ndarray
        :returns: fuel wing weight
        :rtype: ndarray
        """
        return 212.5 / 27.0 * x_shared[5] ** (3.0 / 2.0) / self.math.sqrt(x_shared[3])

    @staticmethod
    def compute_dfuelwing_dar(x_shared):
        """Compute fuel wing weight.

        :param x_shared: system design variables
        :type x_shared: ndarray
        :returns: fuel wing weight
        :rtype: ndarray
        """
        dfuelwing_dar = 212.5 / 27.0 * x_shared[5] ** (3.0 / 2.0)
        dfuelwing_dar *= x_shared[0] * -0.5 * x_shared[3] ** (-3.0 / 2.0)
        return dfuelwing_dar

    def compute_dfuelwing_dsref(self, x_shared):
        """Compute fuel wing weight.

        :param x_shared: system design variables
        :type x_shared: ndarray
        :returns: fuel wing weight
        :rtype: ndarray
        """
        return (
            637.5
            / 54.0
            * x_shared[5] ** 0.5
            * x_shared[0]
            / self.math.sqrt(x_shared[3])
        )

    def __compute_dhalfspan_dar(self, x_shared):
        """Compute partial derivative of half-span wrt aspect ratio.

        :param x_shared: system design variables
        :type x_shared: ndarray
        :returns:  d(half_span)_d(AR)
        :rtype: ndarray
        """
        dadimspan_dspan = self.__compute_dadimspan_dspan(x_shared)
        return (
            dadimspan_dspan
            * x_shared[5]
            / (8.0 * self.base.compute_half_span(x_shared))
        )

    def __compute_dadimspan_dsref(self, x_shared):
        """Compute partial derivative of half-span wrt wing surface.

        :param x_shared: system design variables
        :type x_shared: ndarray
        :returns:  d(half_span)_d(sref)
        :rtype: ndarray
        """
        dadimspan_dspan = self.__compute_dadimspan_dspan(x_shared)
        return (
            dadimspan_dspan
            * x_shared[3]
            / (8.0 * self.base.compute_half_span(x_shared))
        )

    def __compute_dadimspan_dspan(self, x_shared):
        """Compute partial derivative of adim half-span of polynomial function wrt half-
        span.

        :param x_shared: system design variables
        :type x_shared: ndarray
        :returns:  dadimhalf_span_dhalf_span
        :rtype: ndarray
        """
        half_span = self.base.compute_half_span(x_shared)
        return self.base.derive_normalize_s(self.half_span_initial, half_span)

    def __compute_dadimx_dx(self, x_1):
        """Compute partial derivative of adim sectional area of polynomial function wrt
        sectional area.

        :param x_1: local design variables
        :type x_1: ndarray
        :returns:  dadimx_dx
        :rtype: ndarray
        """
        return self.base.derive_normalize_s(self.x_initial, x_1[1])

    def __compute_dadimtaper_dtaper(self, x_shared):
        """Compute partial derivative of adim taper ratio area of polynomial function
        wrt taper ratio.

        :param x_shared: system design variables
        :type x_shared: ndarray
        :returns:  dadimtaper_dtaper
        :rtype: ndarray
        """
        return self.base.derive_normalize_s(self.tc_initial, x_shared[0])

    def __compute_dadimlift_dlift(self, y_21):
        """Compute partial derivative of adim lift  area of polynomial function wrt
        lift.

        :param y_21: coupling design variables (weight)
        :type y_21: ndarray
        :returns:  dadimtaper_dtaper
        :rtype: ndarray
        """
        return self.base.derive_normalize_s(self.lift_initial, y_21[0])

    @staticmethod
    def compute_weight_ratio(y_1):
        """Computation of weight ratio of Breguet formula.

        :param y_1: shared variables coming from blackbox_structure
            - y_1[0]: total aircraft weight
            - y_1[1]: fuel weight
        :type y_1: ndarray
        :returns: Wt / (Wt -Wf)
        :rtype: ndarray
        """
        return y_1[0] / (y_1[0] - y_1[1])

    def blackbox_structure(self, x_shared, y_21, y_31, x_1, true_cstr=False):
        """This function calculates the weight of the aircraft by structure and adds
        them to obtain a total aircraft weight.

        :param x_shared: shared design variable vector:

            - x_shared[0]: thickness/chord ratio
            - x_shared[1]: altitude
            - x_shared[2]: Mach
            - x_shared[3]: aspect ratio
            - x_shared[4]: wing sweep
            - x_shared[5]: wing surface area

        :type x_shared: ndarray
        :param y_21: lift
        :type y_21: ndarray
        :param y_31: engine weight
        :type y_31: ndarray
        :param x_1: weight design variables:

            - x_1[0]: wing taper ratio
            - x_1[1]: wingbox x-sectional area as poly. funct

        :type x_1: ndarray
        :param true_cstr: Default value = False)
        :returns: g_1,y_1, y_11, y_12, y_14:

            - g_1: vector of constraints for weight analysis

                - g_1[0] to g_1[4]: stress on wing
                - g_1[5]: wing twist as constraint

            - y_1: weight analysis outputs

                - y_1[0]: total aircraft weight
                - y_1[1]: fuel weight
                - y_1[2]: wing twist

            - y_12: shared variables used for aero computations
                (blackbox_aerodynamics)

                - y_12[0]: total aircraft weight
                - y_12[1]: wing twist

            - y_14: shared variables used for range computation
                (BlackBoxMission)

                - y_14[0]: total aircraft weight
                - y_14[1]: fuel weight

        :rtype: ndarray, ndarray, ndarray, ndarray
        """
        y_12 = zeros(2, dtype=self.dtype)
        y_14 = zeros(2, dtype=self.dtype)

        y_1, y_11 = self.poly_structure(x_shared, x_1, y_21, y_31)

        g_1 = self.poly_structure_constraints(x_shared, x_1, y_21)
        g_1[5] = y_1[2]

        # Coupling variables
        y_12[1] = y_1[2]
        y_12[0] = y_1[0]
        y_14[0] = y_1[0]
        y_14[1] = y_1[1]
        if not true_cstr:
            g_1 = append(
                g_1[0:5] - self.STRESS_LIMIT,
                (g_1[5] - self.TWIST_UPPER_LIMIT, self.TWIST_LOWER_LIMIT - g_1[5]),
            )

        return y_1, y_11, y_12, y_14, g_1

    def __initialize_jacobian(self):
        """Initialization of jacobian matrix.

        :returns:  jacobian
        :rtype: dict of dict of ndarray
        """
        # Jacobian matrix as a dictionary
        jacobian = {"y_1": {}, "g_1": {}, "y_12": {}, "y_14": {}, "y_11": {}}

        n_y1 = 3
        jacobian["y_1"]["x_shared"] = zeros((n_y1, 6), dtype=self.dtype)
        jacobian["y_1"]["x_1"] = zeros((n_y1, 2), dtype=self.dtype)
        jacobian["y_1"]["y_21"] = zeros((n_y1, 1), dtype=self.dtype)
        jacobian["y_1"]["y_31"] = zeros((n_y1, 1), dtype=self.dtype)

        n_y11 = 1
        jacobian["y_11"]["x_shared"] = zeros((n_y11, 6), dtype=self.dtype)
        jacobian["y_11"]["x_1"] = zeros((n_y11, 2), dtype=self.dtype)
        jacobian["y_11"]["y_21"] = zeros((n_y11, 1), dtype=self.dtype)
        jacobian["y_11"]["y_31"] = zeros((n_y11, 1), dtype=self.dtype)

        n_y12 = 2
        jacobian["y_12"]["x_shared"] = zeros((n_y12, 6), dtype=self.dtype)
        jacobian["y_12"]["x_1"] = zeros((n_y12, 2), dtype=self.dtype)
        jacobian["y_12"]["y_21"] = zeros((n_y12, 1), dtype=self.dtype)
        jacobian["y_12"]["y_31"] = zeros((n_y12, 1), dtype=self.dtype)

        for key, jac_val in jacobian["y_12"].items():
            jacobian["y_14"][key] = jac_val
        return jacobian

    def derive_blackbox_structure(self, x_shared, y_21, y_31, x_1, true_cstr=False):
        """Compute jacobian matrix of structural analysis y_1 is the vector of
        structural outputs and g_1 are the structural constraints:

        - y_1[0]: total aircraft weight
        - y_1[1]: fuel weight
        - y_1[2]: wing twist

        :param x_shared: shared design variable vector
        :type x_shared: ndarray
        :param y_21: coupling variable from aerodynamics (lift)
        :type y_21: ndarray
        :param y_31: coupling variable from propulsion (Engine weight)
        :type y_31: ndarray
        :param x_1: structure design variable vector

            - x_1[0]: wing taper ratio
            - x_1[1]: wingbox x-sectional area as poly. funct

        :type x_1: ndarray
        :param true_cstr: Default value = False)
        :returns: jacobian : Jacobian matrix
        :rtype: dict(dict(ndarray))
        """
        # Jacobian matrix as a dictionary
        jacobian = self.__initialize_jacobian()

        jacobian = self.derive_poly_structure(jacobian, x_shared, x_1, y_21, y_31)

        # Stress constraints
        jacobian = self.derive_constraints(jacobian, x_shared, x_1, y_21, true_cstr)

        # Twist constraints
        jacobian["g_1"]["x_1"][5, :] = jacobian["y_1"]["x_1"][2, :]
        jacobian["g_1"]["x_shared"][5, :] = jacobian["y_1"]["x_shared"][2, :]
        jacobian["g_1"]["y_21"][5, :] = jacobian["y_1"]["y_21"][2, :]
        jacobian["g_1"]["y_31"][5, :] = jacobian["y_1"]["y_31"][2, :]

        if not true_cstr:
            jacobian["g_1"]["x_1"][6, :] = -jacobian["y_1"]["x_1"][2, :]
            jacobian["g_1"]["x_shared"][6, :] = -jacobian["y_1"]["x_shared"][2, :]
            jacobian["g_1"]["y_21"][6, :] = -jacobian["y_1"]["y_21"][2, :]
            jacobian["g_1"]["y_31"][6, :] = -jacobian["y_1"]["y_31"][2, :]

        # Coupling variables
        jacobian = self.__set_coupling_jacobian(jacobian)

        return jacobian

    @staticmethod
    def __set_coupling_jacobian(jacobian):
        """Set jacobian of coupling variables."""
        # Coupling variables
        for der_v, jac_loc in jacobian["y_1"].items():
            jacobian["y_12"][der_v][1, :] = jac_loc[2, :]
            jacobian["y_12"][der_v][0, :] = jac_loc[0, :]
            jacobian["y_14"][der_v] = jac_loc[0:2, :]
        return jacobian

    def derive_poly_structure(self, jacobian, x_shared, x_1, y_21, y_31):
        """Compute derivatives of structural variables from a polynomial approximation.

        :param jacobian: jacobian matrix
        :type jacobian: dict(dict(ndarray))
        :param x_shared: shared design variable vector
        :type x_shared: ndarray
        :param x_1: structure design variable vector:

                - x_1[0]: wing taper ratio
                - x_1[1]: wingbox x-sectional area as poly. funct

        :type x_1: ndarray
        :param y_21: coupling variable from aerodynamics
        :type y_21: ndarray
        :param y_31: coupling variable from propulsion
        :type y_31: ndarray
        :returns: updated jacobian matrix
        :rtype: dict(dict(ndarray))
        """
        half_span = self.base.compute_half_span(x_shared)  # 1/2 span
        aero_center = self.base.compute_aero_center(x_1)
        # wing aero. center location
        y_1, _ = self.poly_structure(x_shared, x_1, y_21, y_31)

        # Derivation of Fuel Weight Wf=y_1[1]
        # dWf/dx=dWfd(lamda= =0
        # jacobian['y_1']['x_1'][1, :] = 0.0  # Fuel weigh does not depend on x_1

        # dWf/d(t/c)
        jacobian["y_1"]["x_shared"][1, 0] = self.compute_dfuelwing_dtoverc(x_shared)
        #        jacobian['y_1']['x_shared'][1, 1] = 0.0  # dWf/dh
        #        jacobian['y_1']['x_shared'][1, 2] = 0.0  # dWf/dM
        # dWf/d(AR)
        jacobian["y_1"]["x_shared"][1, 3] = self.compute_dfuelwing_dar(x_shared)

        #        jacobian['y_1']['x_shared'][1, 4] = 0.0  # dWf/d(sweep)
        # dWf/dsref
        jacobian["y_1"]["x_shared"][1, 5] = self.compute_dfuelwing_dsref(x_shared)
        # dWf_dLift= 0.0
        # dWf_dWe= 0.0

        # Derivation of wing twist = y_1[2]
        s_initial1, s_1, flag1, bound1 = self.__set_coeff_twist(
            x_1, y_21, half_span, aero_center
        )
        _, a_i, a_ij, s_shifted = self.base.derive_poly_approx(
            s_initial1, s_1, flag1, bound1
        )

        dcenter_dlambda = self.__compute_dcenter_dlambda(x_1)
        jacobian["y_1"]["x_1"][2, 0] = dcenter_dlambda * (
            a_i[2]
            + a_ij[2, 0] * s_shifted[0]
            + a_ij[2, 1] * s_shifted[1]
            + a_ij[2, 2] * s_shifted[2]
            + a_ij[2, 3] * s_shifted[3]
        )

        dadimx_dx = self.__compute_dadimx_dx(x_1)
        jacobian["y_1"]["x_1"][2, 1] = dadimx_dx * (
            a_i[0]
            + a_ij[0, 0] * s_shifted[0]
            + a_ij[0, 1] * s_shifted[1]
            + a_ij[0, 2] * s_shifted[2]
            + a_ij[0, 3] * s_shifted[3]
        )
        #        J['y_1']['x_shared'][2, 0] = 0.0  # dtwist_d(t/c)
        #        J['y_1']['x_shared'][2, 1] = 0.0  # dtwist_dh
        #        J['y_1']['x_shared'][2, 2] = 0.0  # dtwist_dM

        #         dLbar_dL = 1.0 / self.L_initial  # Half-span
        dhalfspan_dar = self.__compute_dhalfspan_dar(x_shared)
        jacobian["y_1"]["x_shared"][2, 3] = dhalfspan_dar * (
            a_i[1]
            + a_ij[1, 0] * s_shifted[0]
            + a_ij[1, 1] * s_shifted[1]
            + a_ij[1, 2] * s_shifted[2]
            + a_ij[1, 3] * s_shifted[3]
        )

        #        jacobian['y_1']['x_shared'][2, 4] = 0.0  # dtwistdsweep

        dhalfspan_dsref = self.__compute_dadimspan_dsref(x_shared)
        jacobian["y_1"]["x_shared"][2, 5] = dhalfspan_dsref * (
            a_i[1]
            + a_ij[1, 0] * s_shifted[0]
            + a_ij[1, 1] * s_shifted[1]
            + a_ij[1, 2] * s_shifted[2]
            + a_ij[1, 3] * s_shifted[3]
        )

        dadimlift_dlift = self.__compute_dadimlift_dlift(y_21)
        jacobian["y_1"]["y_21"][2, 0] = dadimlift_dlift * (
            a_i[3]
            + a_ij[0, 3] * s_shifted[0]
            + a_ij[1, 3] * s_shifted[1]
            + a_ij[2, 3] * s_shifted[2]
        )

        #        jacobian['y_1']['y_31'][2, 0] = 0.0  # dtwist_dWengine

        # Derivation of total weight = y_1[0] (requires derivation of wing weight)
        # Calculation of wingbox X-sectional thickness
        wing_w, ww_coeff, _, a_i, a_ij, s_shifted = self.compute_wing_weight(
            x_shared, x_1, y_21, linearize=True
        )

        # dtotal_weight_d(t/c)
        jacobian["y_1"]["x_shared"][0, 0] = (
            -0.4 * x_shared[0] ** (-1.0) * wing_w + jacobian["y_1"]["x_shared"][1, 0]
        )

        # dtotal_weight_dh = 0.0
        # dtotal_weight_dM = 0.0

        dy11_dz = jacobian["y_1"]["x_shared"]
        # dtotal_weight_dAR
        dy11_dz[0, 3] = 0.5 * x_shared[3] ** (-1.0) * wing_w
        dy11_dz[0, 3] += jacobian["y_1"]["x_shared"][1, 3]

        # d1total_weight_dsweep
        dy11_dz[0, 4] = DEG_TO_RAD
        dy11_dz[0, 4] *= self.math.sin(x_shared[4] * DEG_TO_RAD) * wing_w
        dy11_dz[0, 4] /= self.math.cos(x_shared[4] * DEG_TO_RAD)

        # dtotal_weight_dsref
        dy11_dz[0, 5] = 0.749 * x_shared[5] ** (-1.0) * wing_w
        dy11_dz[0, 5] += jacobian["y_1"]["x_shared"][1, 5]

        # dtotal_weight_d(lambda)
        jacobian["y_1"]["x_1"][0, 0] = (
            0.1 * (1 + x_1[0]) ** (-1.0) * wing_w + jacobian["y_1"]["x_1"][1, 0]
        )

        # dtotal_weight_dx
        jacobian["y_1"]["x_1"][0, 1] = ww_coeff * dadimx_dx
        jacobian["y_1"]["x_1"][0, 1] *= a_i[0] + a_ij[0, 0] * s_shifted[0]

        jacobian["y_1"]["x_1"][0, 1] += jacobian["y_1"]["x_1"][1, 1]

        # dtotal_weight_dLift
        jacobian["y_1"]["y_21"][0, 0] = 0.557 * y_21[0] ** -1 * wing_w
        # dtotal_weight_dWe
        jacobian["y_1"]["y_31"][0, 0] = 1.0

        # y_11 = log(y_1[0])- log((y_1[0] - y_1[1]))
        # dy_11 = dy_1[0]/y_1[0] +(dy_1[1]-dy_1[0])/(y_1[0] - y_1[1])
        #        y_11[3] = y_1[0] / (y_1[0] - y_1[1])
        #         y_11 = array([self.math.log( y_1[0] / (y_1[0] - y_1[1]))])
        #     return arr

        for out_v, jac_y_1 in jacobian["y_1"].items():
            val = jac_y_1[0, :] / y_1[0]
            val -= (jac_y_1[0, :] - jac_y_1[1, :]) / (y_1[0] - y_1[1])

            jacobian["y_11"][out_v][0, :] = val
        return jacobian

    def derive_constraints(self, jacobian, x_shared, x_1, y_21, true_cstr=False):
        """Compute derivative structural constraints from a polynomial approximation.

        :param jacobian: Jacobian matrix
        :type jacobian: dict(dict(ndarray))
        :param x_shared: shared design variable vector
        :type x_shared: ndarray
        :param x_1: structure design variable vector:

            - x_1[0]: wing taper ratio
            - x_1[1]: wingbox x-sectional area as poly. funct

        :type x_1: ndarray
        :param y_21: coupling variable from aerodynamics
        :type y_21: ndarray
        :param true_cstr: Default value = False)
        :returns: updated Jacobian matrix J
        :rtype: dict(dict(ndarray))
        """

        if true_cstr is False:
            n_g = 7
        else:
            n_g = 6
        jacobian["g_1"]["x_shared"] = zeros((n_g, 6), dtype=self.dtype)
        jacobian["g_1"]["x_1"] = zeros((n_g, 2), dtype=self.dtype)
        jacobian["g_1"]["y_21"] = zeros((n_g, 1), dtype=self.dtype)
        jacobian["g_1"]["y_31"] = zeros((n_g, 1), dtype=self.dtype)

        half_span = self.base.compute_half_span(x_shared)  # 1/2 span
        aero_center = self.base.compute_aero_center(x_1)

        # Stress constraints
        s_initial, s_new, flag = self.__set_coeff_stress(
            x_1, x_shared, y_21, half_span, aero_center
        )

        ones_mat = ones(5, dtype=self.dtype)
        for i in range(5):
            bound = 0.1 * ones_mat + i * 0.05 * ones_mat
            _, a_i, a_ij, s_shifted = self.base.derive_poly_approx(
                s_initial, s_new, flag, bound
            )
            dg_dx_1, dg_dz, dg_dy_21, dg_dy_31 = self.__der_constraint(
                x_1, x_shared, y_21, a_i, a_ij, s_shifted
            )
            jacobian["g_1"]["x_1"][i, :] = dg_dx_1[:]
            jacobian["g_1"]["x_shared"][i, :] = dg_dz[:]
            jacobian["g_1"]["y_21"][i, :] = dg_dy_21[:]
            jacobian["g_1"]["y_31"][i, :] = dg_dy_31[:]

        return jacobian

    def poly_structure(self, x_shared, x_1, y_21, y_31):
        """Compute structural variables from a polynomial approximation.

        :param x_shared: shared design variable vector
        :type x_shared: ndarray
        :param x_1: structure design variable vector:

            - x_1[0]: wing taper ratio
            - x_1[1]: wingbox x-sectional area as poly. funct

        :type x_1: ndarray
        :param y_21: coupling variable from aerodynamics
        :type y_21: ndarray
        :param y_31: coupling variable from propulsion: engine weight
        :type y_31: ndarray
        :returns: y_1, g_1: structural variables and structural constraints:

            - y_1[0]: total aircraft weight
            - y_1[1]: fuel weight
            - y_1[2]: wing twist

        :rtype: ndarray, ndarray
        """
        y_1 = zeros(3, dtype=self.dtype)
        #         t = self.base.compute_thickness(x_shared)  # wing thickness

        # Calculation of wing twist
        y_1[2] = self.compute_wing_twist(x_shared, x_1, y_21)

        # Calculation of wingbox X-sectional thickness
        wing_w = self.compute_wing_weight(x_shared, x_1, y_21)
        fuel_wing_weight = self.compute_fuelwing_weight(x_shared)
        y_1_i1 = self.constants[0] + fuel_wing_weight
        y_1[1] = y_1_i1  # Fuel weight
        y_1[0] = self.constants[1] + wing_w + y_1_i1 + y_31[0]

        # This is the mass term in the Breguet range equation.
        y_11 = array([self.math.log(self.compute_weight_ratio(y_1))], dtype=self.dtype)

        return y_1, y_11

    def compute_wing_twist(self, x_shared, x_1, y_21):
        """Compute the wing twist (a.k.a. theta, y_12[1], y_1[2]).

        :param x_shared: shared design variables
        :type x_shared: ndarray
        :param x_1: structure design variable
        :type x_1: ndarray
        :param y_21: coupling variable from aerodynamics
        :type y_21: ndarray
        :returns: the wing twist
        :rtype: float
        """
        # Compute the half span and the aerodynamic center
        half_span = self.base.compute_half_span(x_shared)
        aero_center = self.base.compute_aero_center(x_1)

        s_initial1, s_1, flag1, bound1 = self.__set_coeff_twist(
            x_1, y_21, half_span, aero_center
        )
        return self.base.poly_approx(s_initial1, s_1, flag1, bound1)

    def __der_constraint(self, x_1, x_shared, y_21, a_i, a_ij, s_shifted):
        """Compute derivative of a structural constraints from a polynomial
        approximation on a section.

        :param x_1: structure design variable vector:

            - x_1[0]: wing taper ratio
            - x_1[1]: wingbox x-sectional area as poly. funct

        :type x_1: ndarray
        :param x_shared: shared design variable vector
        :type x_shared: ndarray
        :param a_i: linear polynomial coeff
        :type a_i: ndarray
        :param a_ij: quadratic polynomial coeff
        :type a_ij: ndarray
        :param s_shifted: normalized design variables
            (local at polynomial function)
        :type s_shifted: ndarray
        :returns: update Jacobian matrix
        :rtype: dict(dict(ndarrays))
        """
        dcenter_dlambda = self.__compute_dcenter_dlambda(x_1)

        dadimx_dx = self.__compute_dadimx_dx(x_1)

        dadimhalfspan_dar = self.__compute_dhalfspan_dar(x_shared)

        dadimhalfspan_dsref = self.__compute_dadimspan_dsref(x_shared)

        dadimtaper_dtaper = self.__compute_dadimtaper_dtaper(x_shared)

        dadim_lift_dlift = self.__compute_dadimlift_dlift(y_21)

        dg_dz = zeros((1, 6), dtype=self.dtype)
        dg_dx_1 = zeros((1, 2), dtype=self.dtype)
        dg_dy_21 = zeros((1, 1), dtype=self.dtype)
        dg_dy_31 = zeros((1, 1), dtype=self.dtype)
        dg_dx_1[0, 0] = dcenter_dlambda * (
            a_i[4]
            + a_ij[4, 0] * s_shifted[0]
            + a_ij[4, 1] * s_shifted[1]
            + a_ij[4, 2] * s_shifted[2]
            + a_ij[4, 3] * s_shifted[3]
            + a_ij[4, 4] * s_shifted[4]
        )

        # dsigma/dx
        dg_dx_1[0, 1] = dadimx_dx * (
            a_i[2]
            + a_ij[2, 0] * s_shifted[0]
            + a_ij[2, 1] * s_shifted[1]
            + a_ij[2, 2] * s_shifted[2]
            + a_ij[2, 3] * s_shifted[3]
            + a_ij[2, 4] * s_shifted[4]
        )

        # dsigma1/d(t/c)
        dg_dz[0, 0] = dadimtaper_dtaper * (
            a_i[0]
            + a_ij[0, 0] * s_shifted[0]
            + a_ij[0, 1] * s_shifted[1]
            + a_ij[0, 2] * s_shifted[2]
            + a_ij[0, 3] * s_shifted[3]
            + a_ij[0, 4] * s_shifted[4]
        )

        # dsigma1/dh
        dg_dz[0, 1] = 0.0

        # dsigma1/dM
        dg_dz[0, 2] = 0.0

        # dsigma1/dAR
        dg_dz[0, 3] = dadimhalfspan_dar * (
            a_i[3]
            + a_ij[3, 0] * s_shifted[0]
            + a_ij[3, 1] * s_shifted[1]
            + a_ij[3, 2] * s_shifted[2]
            + a_ij[3, 3] * s_shifted[3]
            + a_ij[3, 4] * s_shifted[4]
        )

        # dsigma1/dsweep
        dg_dz[0, 4] = 0.0

        # dsigma1/dsref
        dg_dz[0, 5] = dadimhalfspan_dsref * (
            a_i[3]
            + a_ij[3, 0] * s_shifted[0]
            + a_ij[3, 1] * s_shifted[1]
            + a_ij[3, 2] * s_shifted[2]
            + a_ij[3, 3] * s_shifted[3]
            + a_ij[3, 4] * s_shifted[4]
        )

        # dsigma1/dLift
        dg_dy_21[0, 0] = dadim_lift_dlift * (
            a_i[1]
            + a_ij[1, 0] * s_shifted[0]
            + a_ij[1, 1] * s_shifted[1]
            + a_ij[1, 2] * s_shifted[2]
            + a_ij[1, 3] * s_shifted[3]
            + a_ij[1, 4] * s_shifted[4]
        )

        # dsigma1/dWengine
        dg_dy_31[0, 0] = 0.0

        return dg_dx_1, dg_dz, dg_dy_21, dg_dy_31

    def poly_structure_constraints(self, x_shared, x_1, y_21):
        """Compute structural constraints from a polynomial approximation.

        :param x_shared: shared design variable vector
        :type x_shared: ndarray
        :param x_1: structure design variable vector:

            - x_1[0]: wing taper ratio
            - x_1[1]: wingbox x-sectional area as poly. funct

        :type x_1: ndarray
        :param y_21: coupling variable from aerodynamics
        :type y_21: ndarray
        :returns: g_1: structural constraints
        :rtype: ndarray
        """

        half_span = self.base.compute_half_span(x_shared)  # 1/2 span
        # wing aero. center location
        aero_center = self.base.compute_aero_center(x_1)

        g_1 = zeros(6, dtype=self.dtype)
        s_initial, s_new, flag = self.__set_coeff_stress(
            x_1, x_shared, y_21, half_span, aero_center
        )
        loc_ones = ones(5, dtype=self.dtype)
        for i in range(5):
            bound = 0.1 * loc_ones + i * 0.05 * loc_ones
            g_1[i] = self.base.poly_approx(s_initial, s_new, flag, bound)

        return g_1



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Calibration of a machine learning algorithm.

A machine learning algorithm depends on hyper-parameters,
e.g. the number of clusters for a clustering algorithm,
the regularization constant for a regression model,
the kernel for a Gaussian process regression, ...
Its ability to generalize the information learned during the training stage,
and thus to avoid over-fitting,
which is an over-reliance on the learning data set,
depends on the values of these hyper-parameters.
Thus,
the hyper-parameters minimizing the learning quality measure are rarely
those minimizing the generalization one.
Classically,
the generalization one decreases before growing again as the model becomes more complex,
while the learning error keeps decreasing.
This phenomenon is called the curse of dimensionality.

In this module,
the :class:`.MLAlgoCalibration` class aims to calibrate the hyper-parameters
in order to minimize this measure of the generalization quality
over a calibration parameter space.
This class relies on the :class:`.MLAlgoAssessor` class
which is a discipline (:class:`.MDODiscipline`)
built from a machine learning algorithm (:class:`.MLAlgo`),
a dataset (:class:`.Dataset`),
a quality measure (:class:`.MLQualityMeasure`)
and various options for the data scaling,
the quality measure
and the machine learning algorithm.
The inputs of this discipline are hyper-parameters of the machine learning algorithm
while the output is the quality criterion.
"""

from __future__ import division, unicode_literals

from typing import Dict, Iterable, Optional, Union

from numpy import argmax, argmin, array, ndarray

from gemseo.algos.design_space import DesignSpace
from gemseo.algos.doe.doe_factory import DOEFactory
from gemseo.core.dataset import Dataset
from gemseo.core.discipline import MDODiscipline
from gemseo.core.doe_scenario import DOEScenario
from gemseo.core.mdo_scenario import MDOScenario
from gemseo.core.scenario import ScenarioInputDataType
from gemseo.mlearning.core.factory import MLAlgoFactory
from gemseo.mlearning.core.ml_algo import MLAlgo, MLAlgoParameterType, TransformerType
from gemseo.mlearning.qual_measure.quality_measure import MLQualityMeasure

MeasureOptionsType = Dict[str, Union[bool, int, Dataset]]


class MLAlgoAssessor(MDODiscipline):
    """Discipline assessing the quality of a machine learning algorithm.

    This quality depends on the values of parameters to calibrate
    with the :class:`.MLAlgoCalibration`.

    Attributes:
        algo (str): The name of a machine learning algorithm.
        measure (MLQualityMeasure): The measure
            to assess the machine learning algorithm.
        measure_options (Dict[str,Union[int,Dataset]]): The options
            of the quality measure.
        parameters (List(str)): The parameters of the machine learning algorithm
            to calibrate.
        dataset (Dataset): The learning dataset.
        transformer (TransformerType): The transformation strategy for data groups.
        algos (List(MLAlgo)): The instances of the machine learning algorithm
            (one per execution of the machine learning algorithm assessor).
    """

    CRITERION = "criterion"
    LEARNING = "learning"
    MULTIOUTPUT = "multioutput"

    def __init__(
        self,
        algo,  # type: str
        dataset,  # type: Dataset
        parameters,  # type: Iterable[str]
        measure,  # type: MLQualityMeasure
        measure_options=None,  # type: Optional[MeasureOptionsType]
        transformer=None,  # type: Optional[TransformerType]
        **algo_options  # type: MLAlgoParameterType
    ):  # type: (...) -> None
        """
        Args:
            algo: The name of a machine learning algorithm.
            dataset: A learning dataset.
            parameters: The parameters of the machine learning algorithm to calibrate.
            measure: A measure to assess the machine learning algorithm.
            measure_options: The options of the quality measure.
                If "multioutput" is missing,
                it is added with False as value.
                If None, do not use quality measure options.
            transformer (Dict[str,Transformer]): The strategies
                to transform the variables.
                The values are instances of :class:`.Transformer`
                while the keys are the names of
                either the variables
                or the groups of variables,
                e.g. "inputs" or "outputs" in the case of the regression algorithms.
                If a group is specified,
                the :class:`.Transformer` will be applied
                to all the variables of this group.
                If None, do not transform the variables.
            **options: The options of the machine learning algorithm.

        Raises:
            ValueError: If the measure option "multioutput" is True.
        """
        super(MLAlgoAssessor, self).__init__()
        self.input_grammar.initialize_from_data_names(parameters)
        self.output_grammar.initialize_from_data_names([self.CRITERION, self.LEARNING])
        self.algo = algo
        self.measure = measure
        self.measure_options = measure_options or {}
        self.parameters = algo_options
        self.data = dataset
        self.transformer = transformer
        self.algos = []

        if "multioutput" in measure_options and measure_options["multioutput"]:
            raise ValueError("MLAlgoAssessor does not support multioutput.")
        measure_options[self.MULTIOUTPUT] = False

    def _run(self):  # type: (...) -> None
        """Run method.

        This method creates a new instance of the machine learning algorithm, from the
        hyper-parameters stored in the local_data attribute of the
        :class:`.MLAlgoAssessor`. It trains it on the learning dataset and measures its
        quality with the :class:`.MLQualityMeasure`.
        """
        inputs = self.get_input_data()
        for index in inputs:
            if len(inputs[index]) == 1:
                inputs[index] = inputs[index][0]
        self.parameters.update(inputs)
        factory = MLAlgoFactory()
        algo = factory.create(
            self.algo, data=self.data, transformer=self.transformer, **self.parameters
        )
        algo.learn()
        measure = self.measure(algo)
        learning = measure.evaluate(multioutput=False)
        criterion = measure.evaluate(**self.measure_options)
        self.store_local_data(criterion=array([criterion]), learning=array([learning]))
        self.algos.append(algo)


class MLAlgoCalibration(object):
    """Calibration of a machine learning algorithm.

    Attributes:
        algo_assessor (MLAlgoAssessor): The assessor for the machine learning algorithm.
        calibration_space (DesignSpace): The space defining the calibration variables.
        maximize_objective (bool): If True, seek to maximize the quality measure.
        dataset (Dataset): The learning dataset.
        optimal_parameters (Dict[str,ndarray]): The optimal parameters
            for the machine learning algorithm.
        optimal_criterion (float): The optimal quality measure.
        optimal_algorithm (MLAlgo): The optimal machine learning algorithm.
        scenario (Scenario): The scenario
            used to calibrate the machine learning algorithm.
    """

    def __init__(
        self,
        algo,  # type: str
        dataset,  # type: Dataset
        parameters,  # type: Iterable[str]
        calibration_space,  # type: DesignSpace
        measure,  # type: MLQualityMeasure
        measure_options=None,  # type: Optional[MeasureOptionsType]
        transformer=None,  # type: Optional[TransformerType]
        **algo_options  # type: MLAlgoParameterType
    ):  # type: (...) -> None
        """
        Args:
            algo: The name of a machine learning algorithm.
            dataset: A learning dataset.
            parameters: The parameters of the machine learning algorithm
                to calibrate.
            calibration_space: The space defining the calibration variables.
            measure: A measure to assess the machine learning algorithm.
            measure_options: The options of the quality measure.
                If None, do not use the quality measure options.
            transformer: The transformation strategy for the data groups.
                If None, do not transform data.
            **algo_options: The options of the machine learning algorithm.
        """
        disc = MLAlgoAssessor(
            algo,
            dataset,
            parameters,
            measure,
            measure_options,
            transformer,
            **algo_options
        )
        self.algo_assessor = disc
        self.calibration_space = calibration_space
        self.maximize_objective = not measure.SMALLER_IS_BETTER
        disc.set_cache_policy(disc.MEMORY_FULL_CACHE)
        self.dataset = None
        self.optimal_parameters = None
        self.optimal_criterion = None
        self.optimal_algorithm = None
        self.scenario = None

    def execute(
        self,
        input_data,  # type: ScenarioInputDataType
    ):  # type: (...) -> None
        """Calibrate the machine learning algorithm from a driver.

        The driver can be either a DOE or an optimizer.

        Args:
            input_data: The driver properties.
        """
        doe_factory = DOEFactory()

        if doe_factory.is_available(input_data["algo"]):
            self.scenario = DOEScenario(
                [self.algo_assessor],
                "DisciplinaryOpt",
                self.algo_assessor.CRITERION,
                self.calibration_space,
                maximize_objective=self.maximize_objective,
            )
        else:
            self.scenario = MDOScenario(
                [self.algo_assessor],
                "DisciplinaryOpt",
                self.algo_assessor.CRITERION,
                self.calibration_space,
                maximize_objective=self.maximize_objective,
            )
        self.scenario.disciplines[0].cache.clear()
        self.scenario.execute(input_data)
        x_opt = self.scenario.design_space.get_current_x_dict()
        f_opt = self.scenario.get_optimum().f_opt
        cache = self.scenario.disciplines[0].cache
        self.dataset = cache.export_to_dataset(by_group=False)
        if self.maximize_objective:
            algo_opt = self.algos[
                argmax(self.get_history(self.algo_assessor.CRITERION))
            ]
        else:
            algo_opt = self.algos[
                argmin(self.get_history(self.algo_assessor.CRITERION))
            ]
        self.optimal_parameters = x_opt
        self.optimal_criterion = f_opt
        self.optimal_algorithm = algo_opt

    def get_history(
        self,
        name,  # type: str
    ):  # type: (...) -> ndarray
        """Return the history of a variable.

        Args:
            name: The name of the variable.

        Returns:
            The history of the variable.
        """
        if self.dataset is not None:
            return self.dataset.data[name]

    @property
    def algos(self):  # type: (...) -> MLAlgo
        """The trained machine learning algorithms."""
        return self.scenario.disciplines[0].algos



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Pierre-Jean Barjhoux
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A constraints plot."""
from __future__ import division, unicode_literals

import logging
from typing import Sequence

import numpy as np
from matplotlib import pyplot
from matplotlib.colors import SymLogNorm

from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.post.core.colormaps import PARULA, RG_SEISMIC
from gemseo.post.opt_post_processor import OptPostProcessor
from gemseo.utils.py23_compat import PY2

LOGGER = logging.getLogger(__name__)


class ConstraintsHistory(OptPostProcessor):
    """Plot of the constraint function history with line charts.

    Indicate the violation with color on the background.

    The plot method requires the constraint names to plot.
    It is possible either to save the plot, to show the plot or both.
    """

    def __init__(
        self,
        opt_problem,  # type: OptimizationProblem
    ):  # type: (...) -> None
        super(ConstraintsHistory, self).__init__(opt_problem)
        self.cmap = PARULA  # "viridis"  # "jet"
        self.ineq_cstr_cmap = RG_SEISMIC  # "seismic" "PRGn_r"
        self.eq_cstr_cmap = "seismic"  # "seismic" "PRGn_r"

    def _plot(
        self,
        constraints_list,  # type: Sequence[str]
    ):  # type: (...) -> None
        """
        Args:
            constraints_list: The names of the constraints.
        """
        # retrieve the constraints values
        add_dv = False
        all_constr_names = self.opt_problem.get_constraints_names()

        for func in list(constraints_list):
            if func not in all_constr_names:
                raise ValueError(
                    "Cannot build constraints history plot, "
                    "function {} is not among the constraints names "
                    "or does not exist.".format(func)
                )

        vals, vname, _ = self.database.get_history_array(
            constraints_list, add_dv=add_dv
        )

        # harmonization of tables format because constraints can be vectorial
        # or scalars. *vals.shape[0] = iteration, *vals.shape[1] = cstr values
        vals = np.atleast_3d(vals)
        vals = vals.reshape((vals.shape[0], vals.shape[1] * vals.shape[2]))

        # prepare the main window
        nb_iter = vals.shape[0]

        n_funcs = len(vname)
        nrows = n_funcs // 2
        if 2 * nrows < n_funcs:
            nrows += 1

        fig, axes = pyplot.subplots(
            nrows=nrows, ncols=2, sharex=True, sharey=False, figsize=(11, 11)
        )

        fig.suptitle("Evolution of the constraints " + "w.r.t. iterations", fontsize=14)

        n_subplots = nrows * 2

        x_iter = np.arange(nb_iter)
        y_lim = 0.0
        vmax = 0.0
        # for each subplot
        for values, name, i in zip(vals.T, vname, np.arange(n_subplots)):
            # prepare the graph
            axe = axes.ravel()[i]
            axe.grid(True)
            axe.set_title(name)
            axe.set_xlim([0, nb_iter])
            axe.axhline(y_lim, color="k", linewidth=2)

            # plot values in lines
            axe.plot(x_iter, values)

            # Plot color bars
            cstr_matrix = np.atleast_2d(values)
            cmap = self.ineq_cstr_cmap
            vmax = max(vmax, np.max(np.abs(cstr_matrix)))
            extent = -0.5, nb_iter - 0.5, np.min(cstr_matrix), np.max(cstr_matrix)

            # On python 2, base is not defined as a parameter in SymLogNorm()
            if PY2:
                norm = SymLogNorm(linthresh=1.0, vmin=-vmax, vmax=vmax)
            else:
                norm = SymLogNorm(linthresh=1.0, vmin=-vmax, vmax=vmax, base=np.e)

            axe.imshow(
                cstr_matrix,
                cmap=cmap,
                interpolation="nearest",
                aspect="auto",
                extent=extent,
                norm=norm,
                alpha=0.6,
            )

            # plot vertical line the last time that g(x)=0
            indices = np.where(np.diff(np.sign(values)))[0]
            if indices.size != 0:
                ind = indices[-1]
                x_lim = np.interp(y_lim, values[ind : ind + 2], x_iter[ind : ind + 2])
                axe.axvline(x_lim, color="k", linewidth=2)

        self._add_figure(fig)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                        documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Quantification of robustness of the optimum to variables perturbations."""
from __future__ import division, unicode_literals

import numpy as np
from numpy.random import multivariate_normal

from gemseo.post.core.hessians import BFGSApprox, LSTSQApprox, SR1Approx


class RobustnessQuantifier(object):
    """classdocs."""

    AVAILABLE_APPROXIMATIONS = ["BFGS", "SR1", "LEAST_SQUARES"]

    def __init__(self, history, approximation_method="SR1"):
        """Constructor.

        :param history: an approximation history.
        :param approximation_method: an approximation method for the Hessian.
        """
        self.history = history
        if approximation_method not in self.AVAILABLE_APPROXIMATIONS:
            raise ValueError(
                "Unknown hessian approximation method "
                + str(approximation_method)
                + " the available ones are:"
                + str(self.AVAILABLE_APPROXIMATIONS)
            )
        if approximation_method == "SR1":
            approx_class = SR1Approx
        elif approximation_method == "BFGS":
            approx_class = BFGSApprox
        elif approximation_method == "LEAST_SQUARES":
            approx_class = LSTSQApprox
        self.approximator = approx_class(history)
        self.b_mat = None
        self.x_ref = None
        self.f_ref = None
        self.fgrad_ref = None

    def compute_approximation(
        self,
        funcname,
        first_iter=0,
        last_iter=0,
        b0_mat=None,
        at_most_niter=-1,
        func_index=None,
    ):
        """Builds the BFGS approximation for the hessian.

        :param at_most_niter: maximum number of iterations to take
            (Default value = -1).
        :param funcname: param first_iter:  (Default value = 0).
        :param last_iter: Default value = 0).
        :param b0_mat: Default value = None).
        :param func_index: Default value = None).
        :param first_iter:  (Default value = 0).
        """

        self.b_mat, _, x_ref, grad_ref = self.approximator.build_approximation(
            funcname=funcname,
            first_iter=first_iter,
            last_iter=last_iter,
            b_mat0=b0_mat,
            at_most_niter=at_most_niter,
            return_x_grad=True,
            func_index=func_index,
        )
        self.x_ref = x_ref
        self.f_ref = self.approximator.f_ref
        self.fgrad_ref = grad_ref
        return self.b_mat

    def compute_expected_value(self, expect, cov):
        """computes 1/2*E(e.T B e) , where e is a vector of expected values expect and
        covariance matrix cov.

        :param expect: the expected value of inputs.
        :param cov: the covariance matrix of inputs.
        """
        n_vars = len(expect)
        if cov.shape != (n_vars, n_vars):
            raise ValueError("Inconsistent expect and covariance matrices shapes")
        if self.b_mat is None:
            raise ValueError(
                "Build Hessian approximation before"
                + " computing expected_value_offset"
            )
        b_approx = 0.5 * self.b_mat
        exp_val = np.trace(np.dot(b_approx, cov))
        exp_val += np.linalg.multi_dot(
            ((expect - self.x_ref).T, b_approx, (expect - self.x_ref))
        )
        return exp_val

    def compute_variance(self, expect, cov):
        """computes 1/2*E(e.T B e), where e is a vector of expected values expect and
        covariance matrix cov.

        :param expect: the expected value of inputs.
        :param cov: the covariance matrix of inputs.
        """
        if self.b_mat is None:
            raise ValueError(
                "Build Hessian approximation" + " before computing variance"
            )
        n_vars = len(expect)
        if cov.shape != (n_vars, n_vars):
            raise ValueError("Inconsistent expect and covariance matrices shapes")
        b_approx = 0.5 * self.b_mat
        mu_cent = expect - self.x_ref
        v_mat = np.linalg.multi_dot((b_approx, cov, b_approx, cov))
        v_mat += 4 * np.linalg.multi_dot((mu_cent.T, b_approx, cov, b_approx, mu_cent))
        var = 2 * np.trace(v_mat)
        return var

    def compute_function_approximation(self, x_vars):
        """Computes a second order approximation of the function.

        :param x_vars: the point on which the approximation is evaluated.
        :param x_vars: x vars.
        """
        if self.b_mat is None or self.x_ref is None:
            raise ValueError(
                "Build Hessian approximation before"
                + " computing function approximation"
            )
        x_l = x_vars - self.x_ref
        return (
            0.5 * np.linalg.multi_dot((x_l.T, self.b_mat, x_l))
            + np.dot(self.fgrad_ref.T, x_l)
            + self.f_ref
        )

    def compute_gradient_approximation(self, x_vars):
        """Computes a first order approximation of the gradient based on the hessian.

        :param x_vars: the point on which the approximation is evaluated.
        """
        if self.b_mat is None or self.fgrad_ref is None:
            raise ValueError(
                "Build Hessian approximation before"
                + " computing function approximation"
            )
        x_l = x_vars - self.x_ref
        return np.dot(self.b_mat, x_l) + self.fgrad_ref

    def montecarlo_average_var(self, mean, cov, n_samples=100000, func=None):
        """Computes the variance and expected value using Monte Carlo approach.

        :param mean: the mean value.
        :param cov: the covariance matrix.
        :param n_samples: the number of samples for the distribution
            (Default value = 100000).
        :param func: if None, the compute_function_approximation function,
                        otherwise a user function (Default value = None).
        """
        n_dv = len(mean)
        if not cov.shape == (n_dv, n_dv):
            raise ValueError(
                "Covariance matrix dimension " + "incompatible with mean dimensions"
            )
        ran = multivariate_normal(mean=mean, cov=cov, size=n_samples).T
        vals = np.zeros(n_samples)
        if func is None:
            func = self.compute_function_approximation
        for i in range(n_samples):
            vals[i] = func(ran[:, i])
        average = np.average(vals)
        var = np.var(vals)
        return average, var



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""scipy.optimize global optimization library wrapper."""
from __future__ import division, unicode_literals

import logging

from numpy import isfinite, real
from scipy import optimize

from gemseo.algos.opt.opt_lib import OptimizationLibrary

LOGGER = logging.getLogger(__name__)


class ScipyGlobalOpt(OptimizationLibrary):
    """Scipy optimization library interface.

    See OptimizationLibrary.
    """

    LIB_COMPUTE_GRAD = True

    def __init__(self):
        """Constructor.

        Generate the library dict, contains the list
        of algorithms with their characteristics:

        - does it require gradient
        - does it handle equality constraints
        - does it handle inequality constraints
        """
        super(ScipyGlobalOpt, self).__init__()
        doc = "https://docs.scipy.org/doc/scipy/reference/generated/"
        self.lib_dict = {
            "DUAL_ANNEALING": {
                self.INTERNAL_NAME: "dual_annealing",
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: False,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "Dual annealing",
                self.WEBSITE: doc + "scipy.optimize.dual_annealing.html",
            },
            "SHGO": {
                self.INTERNAL_NAME: "shgo",
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: False,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "Simplicial homology global optimization",
                self.WEBSITE: doc + "scipy.optimize.shgo.html",
            },
            "DIFFERENTIAL_EVOLUTION": {
                self.INTERNAL_NAME: "differential_evolution",
                self.REQUIRE_GRAD: False,
                self.POSITIVE_CONSTRAINTS: False,
                self.HANDLE_EQ_CONS: False,
                self.HANDLE_INEQ_CONS: False,
                self.DESCRIPTION: "Differential Evolution algorithm",
                self.WEBSITE: doc + "scipy.optimize.differential_evolution.html",
            },
        }

    def _get_options(
        self,
        max_iter=999,
        ftol_rel=1e-9,
        ftol_abs=1e-9,
        xtol_rel=1e-9,
        xtol_abs=1e-9,
        workers=1,
        updating="immediate",
        atol=0,
        init="latinhypercube",
        recombination=0.7,
        tol=0.01,
        popsize=15,
        strategy="best1bin",
        sampling_method="simplicial",
        niters=1,
        n=100,
        seed=1,
        polish=True,
        **kwargs
    ):  # pylint: disable=W0221
        r"""Sets the options default values

        To get the best and up to date information about algorithms options,
        go to scipy.optimize documentation:
        https://docs.scipy.org/doc/scipy/reference/tutorial/optimize.html

        :param max_iter: maximum number of iterations, ie unique calls to f(x)
        :type max_iter: int
        :param ftol_rel: stop criteria, relative tolerance on the
               objective function,
               if abs(f(xk)-f(xk+1))/abs(f(xk))<= ftol_rel: stop
               (Default value = 1e-9)
        :type ftol_rel: float
        :param ftol_abs: stop criteria, absolute tolerance on the objective
               function, if abs(f(xk)-f(xk+1))<= ftol_rel: stop
               (Default value = 1e-9)
        :type ftol_abs: float
        :param xtol_rel: stop criteria, relative tolerance on the
               design variables,
               if norm(xk-xk+1)/norm(xk)<= xtol_rel: stop
               (Default value = 1e-9)
        :type xtol_rel: float
        :param xtol_abs: stop criteria, absolute tolerance on the
               design variables,
               if norm(xk-xk+1)<= xtol_abs: stop
               (Default value = 1e-9)
        :type xtol_abs: float
        :param kwargs: other algorithms options
        :type kwargs: kwargs
        """
        popts = self._process_options(
            max_iter=max_iter,
            ftol_rel=ftol_rel,
            ftol_abs=ftol_abs,
            xtol_rel=xtol_rel,
            xtol_abs=xtol_abs,
            workers=workers,
            updating=updating,
            atol=atol,
            init=init,
            recombination=recombination,
            seed=seed,
            tol=tol,
            popsize=popsize,
            strategy=strategy,
            sampling_method=sampling_method,
            niters=niters,
            n=n,
            polish=polish,
            **kwargs
        )
        return popts

    def _run(self, **options):
        """Runs the algorithm, to be overloaded by subclasses.

        :param options: the options dict for the algorithm
        """
        # remove normalization from options for algo
        normalize_ds = options.pop(self.NORMALIZE_DESIGN_SPACE_OPTION, True)
        # Get the normalized bounds:
        _, l_b, u_b = self.get_x0_and_bounds_vects(normalize_ds)
        # Replace infinite values with None:
        l_b = [val if isfinite(val) else None for val in l_b]
        u_b = [val if isfinite(val) else None for val in u_b]
        bounds = list(zip(l_b, u_b))

        def real_part_fun(x_vect):
            """Wraps the function and returns the real part."""
            return real(self.problem.objective.func(x_vect))

        fun = real_part_fun

        if self.internal_algo_name == "dual_annealing":
            opt_result = optimize.dual_annealing(
                func=fun,
                bounds=bounds,
                maxiter=10000000,
                local_search_options={},
                initial_temp=5230.0,
                restart_temp_ratio=2e-05,
                visit=2.62,
                accept=-5.0,
                maxfun=10000000.0,
                seed=1,
                no_local_search=False,
                callback=None,
                x0=None,
            )
        elif self.internal_algo_name == "shgo":
            opt_result = optimize.shgo(
                func=fun,
                bounds=bounds,
                args=(),
                constraints=None,
                n=options["n"],
                iters=options["iters"],
                callback=None,
                minimizer_kwargs=None,
                options=None,
                sampling_method=options["sampling_method"],
            )
        elif self.internal_algo_name == "differential_evolution":
            opt_result = optimize.differential_evolution(
                func=fun,
                bounds=bounds,
                args=(),
                strategy=options["strategy"],
                maxiter=10000000,
                popsize=options["popsize"],
                tol=options["tol"],
                mutation=options.get("mutation", (0.5, 1)),
                recombination=options["recombination"],
                seed=options["seed"],
                polish=options["polish"],
                init=options["init"],
                atol=options["atol"],
                updating=options["updating"],
                workers=options["workers"],
            )
        return self.get_optimum_from_database(opt_result.message, opt_result.success)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The Kernel Principal Component Analysis (KPCA) to reduce the dimension of a variable.

The :class:`KPCA` class implements the KCPA wraps the KPCA from Scikit-learn.

Dependence
----------
This dimension reduction algorithm relies on the PCA class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.decomposition.PCA.html>`_.
"""
from __future__ import division, unicode_literals

from typing import Optional, Union

from numpy import ndarray
from sklearn.decomposition import KernelPCA

from gemseo.mlearning.transform.dimension_reduction.dimension_reduction import (
    DimensionReduction,
)
from gemseo.mlearning.transform.transformer import TransformerFitOptionType


class KPCA(DimensionReduction):
    """Kernel principal component dimension reduction algorithm."""

    def __init__(
        self,
        name="KPCA",  # type: str
        n_components=5,  # type: int
        fit_inverse_transform=True,  # type: bool
        kernel="linear",  # type: str
        **parameters  # type: Optional[Union[float,int,str]]
    ):
        """
        Args:
            fit_inverse_transform: If True, learn the inverse transform
                for non-precomputed kernels.
            kernel: The name of the kernel,
                either 'linear', 'poly', 'rbf', 'sigmoid', 'cosine' or 'precomputed'.
            **parameters: The optional parameters for sklearn KPCA constructor.
        """
        super(KPCA, self).__init__(name, n_components=n_components, **parameters)
        self.algo = KernelPCA(
            n_components,
            fit_inverse_transform=fit_inverse_transform,
            kernel=kernel,
            **parameters
        )

    def fit(
        self,
        data,  # type: ndarray
        *args  # type: TransformerFitOptionType
    ):  # type: (...) -> None
        self.algo.fit(data)

    def transform(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        return self.algo.transform(data)

    def inverse_transform(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        return self.algo.inverse_transform(data)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Here is the baseclass to measure the error of machine learning algorithms.

The concept of error measure is implemented with the :class:`.MLErrorMeasure` class and
proposes different evaluation methods.
"""
from __future__ import division, unicode_literals

from typing import List, NoReturn, Optional, Union

from numpy import arange, array_split
from numpy import delete as npdelete
from numpy import ndarray, unique
from numpy.random import choice

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.supervised import MLSupervisedAlgo
from gemseo.mlearning.qual_measure.quality_measure import MLQualityMeasure


class MLErrorMeasure(MLQualityMeasure):
    """An abstract error measure for machine learning."""

    def __init__(
        self,
        algo,  # type: MLSupervisedAlgo
    ):  # type: (...) -> None
        """
        Args:
            algo: A machine learning algorithm for supervised learning.
        """
        super(MLErrorMeasure, self).__init__(algo)

    def evaluate_learn(
        self,
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        samples = self._assure_samples(samples)
        self.algo.learn(samples)
        inputs = self.algo.input_data[samples]
        outputs = self.algo.output_data[samples]
        predictions = self.algo.predict(inputs)
        measure = self._compute_measure(outputs, predictions, multioutput)
        return measure

    def evaluate_test(
        self,
        test_data,  # type:Dataset
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        samples = self._assure_samples(samples)
        self.algo.learn(samples)
        in_grp = test_data.INPUT_GROUP
        out_grp = test_data.OUTPUT_GROUP
        inputs = test_data.get_data_by_group(in_grp)
        outputs = test_data.get_data_by_group(out_grp)
        predictions = self.algo.predict(inputs)
        measure = self._compute_measure(outputs, predictions, multioutput)
        return measure

    def evaluate_kfolds(
        self,
        n_folds=5,  # type: int
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        samples = self._assure_samples(samples)
        inds = samples
        folds = array_split(inds, n_folds)

        in_grp = self.algo.learning_set.INPUT_GROUP
        out_grp = self.algo.learning_set.OUTPUT_GROUP
        inputs = self.algo.learning_set.get_data_by_group(in_grp)
        outputs = self.algo.learning_set.get_data_by_group(out_grp)

        qualities = []
        for n_fold in range(n_folds):
            fold = folds[n_fold]
            train = npdelete(inds, fold)
            self.algo.learn(samples=train)
            expected = outputs[fold]
            predicted = self.algo.predict(inputs[fold])
            quality = self._compute_measure(expected, predicted, multioutput)
            qualities.append(quality)

        quality = sum(qualities) / len(qualities)

        return quality

    def evaluate_bootstrap(
        self,
        n_replicates=100,  # type: int
        samples=None,  # type: Optional[List[int]]
        multioutput=True,  # type: bool
    ):  # type: (...) -> Union[float,ndarray]
        samples = self._assure_samples(samples)
        if isinstance(samples, list):
            n_samples = len(samples)
        else:
            n_samples = samples.size
        inds = arange(n_samples)

        in_grp = self.algo.learning_set.INPUT_GROUP
        out_grp = self.algo.learning_set.OUTPUT_GROUP
        inputs = self.algo.learning_set.get_data_by_group(in_grp)
        outputs = self.algo.learning_set.get_data_by_group(out_grp)

        qualities = []
        for _ in range(n_replicates):
            train = unique(choice(n_samples, n_samples))
            test = npdelete(inds, train)
            self.algo.learn([samples[index] for index in train])
            test_samples = [samples[index] for index in test]
            expected = outputs[test_samples]
            predicted = self.algo.predict(inputs[test_samples])
            quality = self._compute_measure(expected, predicted, multioutput)
            qualities.append(quality)

        quality = sum(qualities) / len(qualities)

        return quality

    def _compute_measure(
        self,
        outputs,  # type: ndarray
        predictions,  # type: ndarray
        multioutput=True,  # type: bool
    ):  # type: (...) -> NoReturn
        """Compute the quality measure.

        Args:
            outputs: The reference data.
            predictions: The predicted labels.
            multioutput: If True, return the quality measure for each
                output component. Otherwise, average these measures.

        Returns:
            The value of the quality measure.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Quadratic approximations of functions from the optimization history."""

from __future__ import division, unicode_literals

import logging
from math import ceil
from typing import Optional

import numpy as np
import pylab
from matplotlib import pyplot
from matplotlib.colors import SymLogNorm
from matplotlib.figure import Figure
from matplotlib.ticker import LogFormatter
from numpy import ndarray
from pylab import plt

from gemseo.algos.opt_problem import OptimizationProblem
from gemseo.post.core.colormaps import PARULA
from gemseo.post.core.hessians import SR1Approx
from gemseo.post.opt_post_processor import OptPostProcessor
from gemseo.utils.py23_compat import PY2

LOGGER = logging.getLogger(__name__)


class QuadApprox(OptPostProcessor):
    """Quadratic approximation of a function.

    And cuts of the approximation.

    The function index can be passed as option.
    """

    SR1_APPROX = "SR1"

    def __init__(
        self,
        opt_problem,  # type: OptimizationProblem
    ):  # type: (...) -> None
        super(QuadApprox, self).__init__(opt_problem)
        self.grad_opt = None

    def _plot(
        self,
        function,  # type: str
        func_index=None,  # type: Optional[int]
    ):  # type: (...) -> None
        """Build the plot and save it.

        Args:
            function: The function name to build the quadratic approximation.
            func_index: The index of the output of interest
                to be defined if the function has a multidimensional output.
                If None and if the output is multidimensional, an error is raised.
        """
        b_mat = self.__build_approx(function, func_index)
        self.out_data_dict["b_mat"] = b_mat
        fig = self.__plot_hessian(b_mat, function)
        self._add_figure(fig, "hess_approx")
        fig = self.__plot_variations(b_mat)
        self._add_figure(fig, "quad_approx")

    def __build_approx(
        self,
        function,  # type: str
        func_index,  # type: int
    ):  # type: (...) -> ndarray
        """Build the approximation.

        Args:
            function: The function name to build the quadratic approximation.
            func_index: The index of the output of interest
                to be defined if the function has a multidimensional output.

        Returns:
             The approximation.
        """
        apprx = SR1Approx(self.database)
        n_x = self.opt_problem.get_dimension()
        max_i = int(1.5 * n_x)
        # Avoid using alpha scaling for hessian otherwise diagonal is messy
        out = apprx.build_approximation(
            function, at_most_niter=max_i, return_x_grad=True, func_index=func_index
        )
        b_mat = out[0]
        grad_opt = out[-1]

        self.grad_opt = grad_opt
        return b_mat

    @staticmethod
    def __plot_hessian(
        hessian,  # type: ndarray
        function,  # type: str
    ):  # type: (...) -> Figure
        """Plot the Hessian of the function.

        Args:
            hessian: The Hessian of the function.
            function: The function name.

        Returns:
            The plot of the Hessian of the function.
        """
        fig = pylab.figure()
        pylab.plt.xlabel(r"$x_i$", fontsize=16)
        pylab.plt.ylabel(r"$x_j$", fontsize=16)
        vmax = max(abs(np.max(hessian)), abs(np.min(hessian)))
        linthresh = 10 ** (np.log10(vmax) - 5.0)

        # On python 2, base is not defined as a parameter in SymLogNorm()
        if PY2:
            norm = SymLogNorm(linthresh=linthresh, vmin=-vmax, vmax=vmax)
        else:
            norm = SymLogNorm(linthresh=linthresh, vmin=-vmax, vmax=vmax, base=np.e)

        # SymLog is a symmetric log scale adapted to negative values
        pylab.imshow(
            hessian,
            cmap=PARULA,
            interpolation="nearest",
            norm=norm,
        )
        pylab.grid(True)
        l_f = LogFormatter(base=10, labelOnlyBase=False)
        thick_min = int(np.log10(linthresh))
        thick_max = int(np.log10(vmax))
        thick_num = thick_max - thick_min + 1
        lvls_pos = np.logspace(thick_min, thick_max, num=thick_num, base=10.0)
        levels_neg = np.sort(-lvls_pos)
        levels = np.concatenate((levels_neg, lvls_pos))
        pylab.plt.colorbar(ticks=levels, format=l_f)
        fig.suptitle("Hessian matrix SR1 approximation of " + str(function))
        return fig

    @staticmethod
    def unnormalize_vector(
        xn_array,  # type: ndarray
        ivar,  # type: int
        lower_bounds,  # type: ndarray
        upper_bounds,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Unormalize a variable with respect to bounds.

        Args:
            xn_array: The normalized variable.
            ivar: The index of the variable bound.
            lower_bounds: The lower bounds of the variable.
            upper_bounds: The upper bounds of the variable.

        Returns:
            The unnormalized variable.
        """
        l_b = lower_bounds[ivar]
        u_b = upper_bounds[ivar]
        return (u_b - l_b) * xn_array * 0.5 + 0.5 * (l_b + u_b)

    def __plot_variations(
        self,
        hessian,  # type: ndarray
    ):  # type: (...) -> Figure
        """Plot the variation plot of the function w.r.t. all variables.

        Args:
            hessian: The Hessian of the function.

        Returns:
            The plot of the variation of the function w.r.t. all variables.
        """
        ndv = hessian.shape[0]
        ncols = int(np.sqrt(ndv)) + 1
        nrows = int(ceil(float(ndv) / ncols))

        xn_vars = np.arange(-1.0, 1.0, 0.01)
        lower_bounds = self.opt_problem.design_space.get_lower_bounds()
        upper_bounds = self.opt_problem.design_space.get_upper_bounds()
        fig = plt.figure()

        for i in range(ndv):
            ax_i = plt.subplot(nrows, ncols, i + 1)
            f_vals = xn_vars ** 2 * hessian[i, i] + self.grad_opt[i] * xn_vars
            self.out_data_dict[i] = f_vals

            x_vars = self.unnormalize_vector(xn_vars, i, lower_bounds, upper_bounds)
            ax_i.plot(x_vars, f_vals, "-", lw=2)

            start, stop = ax_i.get_xlim()
            ax_i.xaxis.set_ticks(np.arange(start, stop, 0.4999999 * (stop - start)))
            start, stop = ax_i.get_ylim()
            ax_i.yaxis.set_ticks(np.arange(start, stop, 0.24999999 * (stop - start)))
            ax_i.set_xlabel(r"$x_{" + str(i) + "}$", fontsize=14)
        pyplot.tight_layout()
        return fig



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Class for the parametric estimation of statistics from a dataset.

Overview
--------

The :class:`.ParametricStatistics` class inherits
from the abstract :class:`.Statistics` class
and aims to estimate statistics from a :class:`.Dataset`,
based on candidate parametric distributions calibrated from this :class:`.Dataset`.

For each variable,

1. the parameters of these distributions are calibrated from the :class:`.Dataset`,
2. the fitted parametric :class:`.Distribution` which is optimal
   in the sense of a goodness-of-fit criterion and a selection criterion
   is selected to estimate the statistics related to this variable.

The :class:`.ParametricStatistics` relies on the OpenTURNS library
through the :class:`.OTDistribution` and :class:`.OTDistributionFitter` classes.

Construction
------------

The :class:`.ParametricStatistics` is built from two mandatory arguments:

- a dataset,
- a list of distributions names,

and can consider optional arguments:

- a subset of variables names
  (by default, statistics are computed for all variables),
- a fitting criterion name
  (by default, BIC is used;
  see :attr:`.AVAILABLE_CRITERIA`
  and :attr:`.AVAILABLE_SIGNIFICANCE_TESTS`
  for more information),
- a level associated with the fitting criterion,
- a selection criterion:

  - 'best':
    select the distribution minimizing
    (or maximizing, depending on the criterion)
    the criterion,
  - 'first':
    select the first distribution
    for which the criterion is greater
    (or lower, depending on the criterion)
    than the level,

- a name for the :class:`.ParametricStatistics` object
  (by default, the name is the concatenation of 'ParametricStatistics'
  and the name of the :class:`.Dataset`).

Capabilities
------------

By inheritance,
a :class:`.ParametricStatistics` object has
the same capabilities as :class:`.Statistics`.
Additional ones are:

- :meth:`.get_fitting_matrix`:
  this method displays the values of the fitting criterion
  for the different variables
  and candidate probability distributions
  as well as the select probability distribution,
- :meth:`.plot_criteria`:
  this method plots the criterion values for a given variable.
"""

from __future__ import division, unicode_literals

import logging
import os
from typing import Dict, Iterable, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
from numpy import array, linspace, ndarray

from gemseo.core.dataset import Dataset
from gemseo.third_party.prettytable.prettytable import PrettyTable
from gemseo.uncertainty.distributions.openturns.distribution import OTDistribution
from gemseo.uncertainty.distributions.openturns.fitting import (
    MeasureType,
    OTDistributionFitter,
)
from gemseo.uncertainty.statistics.statistics import Statistics
from gemseo.uncertainty.statistics.tolerance_interval.distribution import (
    ToleranceIntervalFactory,
    ToleranceIntervalSide,
)

LOGGER = logging.getLogger(__name__)


class ParametricStatistics(Statistics):
    """Parametric estimation of statistics.

    Attributes:
        fitting_criterion (str): The name of the goodness-of-fit criterion,
            measuring how the distribution fits the data.
        level (float): The test level,
            i.e. risk of committing a Type 1 error,
            that is an incorrect rejection of a true null hypothesis,
            for criteria based on test hypothesis.
        selection_criterion (str): The name of the selection criterion
            to select a distribution from a list of candidates.
        distributions (dict(str, dict(str, OTDistribution))): The probability
            distributions of the random variables.

    Examples:
        >>> from gemseo.api import (
        ...     create_discipline,
        ...     create_parameter_space,
        ...     create_scenario
        ... )
        >>> from gemseo.uncertainty.statistics.parametric import ParametricStatistics
        >>>
        >>> expressions = {"y1": "x1+2*x2", "y2": "x1-3*x2"}
        >>> discipline = create_discipline(
        ...     "AnalyticDiscipline", expressions_dict=expressions
        ... )
        >>> discipline.set_cache_policy(discipline.MEMORY_FULL_CACHE)
        >>>
        >>> parameter_space = create_parameter_space()
        >>> parameter_space.add_random_variable(
        ...     "x1", "OTUniformDistribution", minimum=-1, maximum=1
        ... )
        >>> parameter_space.add_random_variable(
        ...     "x2", "OTNormalDistribution", mu=0.5, sigma=2
        ... )
        >>>
        >>> scenario = create_scenario(
        ...     [discipline],
        ...     "DisciplinaryOpt",
        ...     "y1", parameter_space, scenario_type="DOE"
        ... )
        >>> scenario.execute({'algo': 'OT_MONTE_CARLO', 'n_samples': 100})
        >>>
        >>> dataset = discipline.cache.export_to_dataset()
        >>>
        >>> statistics = ParametricStatistics(
        ...     dataset, ['Normal', 'Uniform', 'Triangular']
        ... )
        >>> fitting_matrix = statistics.get_fitting_matrix()
        >>> mean = statistics.mean()
    """

    AVAILABLE_DISTRIBUTIONS = sorted(
        OTDistributionFitter._AVAILABLE_DISTRIBUTIONS.keys()
    )

    AVAILABLE_CRITERIA = sorted(OTDistributionFitter._AVAILABLE_FITTING_TESTS.keys())
    AVAILABLE_SIGNIFICANCE_TESTS = sorted(OTDistributionFitter.SIGNIFICANCE_TESTS)

    def __init__(
        self,
        dataset,  # type: Dataset
        distributions,  # type: Sequence[str]
        variables_names=None,  # type: Optional[Iterable[str]]
        fitting_criterion="BIC",  # type: str
        level=0.05,  # type: float
        selection_criterion="best",  # type: str
        name=None,  # type: Optional[str]
    ):  # type: (...) -> None  # noqa: D205,D212,D415
        """
        Args:
            distributions: The names of the distributions.
            fitting_criterion: The name of
                the goodness-of-fit criterion,
                measuring how the distribution fits the data.
                Use :meth:`.ParametricStatistics.get_criteria`
                to get the available criteria.
            level: A test level,
                i.e. the risk of committing a Type 1 error,
                that is an incorrect rejection of a true null hypothesis,
                for criteria based on test hypothesis.
            selection_criterion: The name of the selection criterion
                to select a distribution from a list of candidates.
                Either 'first' or 'best'.
        """
        super(ParametricStatistics, self).__init__(dataset, variables_names, name)
        significance_tests = OTDistributionFitter.SIGNIFICANCE_TESTS
        self.fitting_criterion = fitting_criterion
        self.selection_criterion = selection_criterion
        LOGGER.info("| Set goodness-of-fit criterion: %s.", fitting_criterion)
        if self.fitting_criterion in significance_tests:
            self.level = level
            LOGGER.info("| Set significance level of hypothesis test: %s.", level)
        else:
            self.level = None
        self._all_distributions = None
        self.distributions = None
        self._build_distributions(distributions)

    def _build_distributions(
        self,
        distributions,  # type: Sequence[str]
    ):  # type: (...) -> None
        """Build distributions from distributions names.

        Args:
            distributions: The names of the distributions.
        """
        self._all_distributions = self._fit_distributions(distributions)
        self.distributions = self._select_best_distributions(distributions)

    def get_fitting_matrix(self):  # type: (...) -> str
        """Get the fitting matrix.

        This matrix contains goodness-of-fit measures
        for each pair < variable, distribution >.

        Returns:
            The printable fitting matrix.
        """
        variables = sorted(self._all_distributions.keys())
        distributions = list(self._all_distributions[variables[0]].keys())
        table = PrettyTable(["Variable"] + distributions + ["Selection"])
        for variable in variables:
            row, _ = self.get_criteria(variable)
            row = [variable] + [row[distribution] for distribution in distributions]
            row = row + [self.distributions[variable]["name"]]
            table.add_row(row)
        return str(table)

    def get_criteria(
        self,
        variable,  # type:str
    ):  # type: (...) -> Tuple[Dict[str,float],bool]
        """Get criteria for a given variable name and the different distributions.

        Args:
            variable: The name of the variable.

        Returns:
            The criterion for the different distributions.
            and an indicator equal to True is the criterion is a p-value.
        """
        all_distributions = self._all_distributions[variable]
        criteria = {
            distribution: result["criterion"]
            for distribution, result in all_distributions.items()
        }
        is_p_value = False
        significance_tests = OTDistributionFitter.SIGNIFICANCE_TESTS
        if self.fitting_criterion in significance_tests:
            criteria = {
                distribution: result[1]["p-value"]
                for distribution, result in criteria.items()
            }
            is_p_value = True
        return criteria, is_p_value

    def plot_criteria(
        self,
        variable,  # type: str
        title=None,  # type: Optional[str]
        save=False,  # type:bool
        show=True,  # type: bool
        n_legend_cols=4,  # type: int
        directory=".",  # type:str
    ):  # type: (...) -> None
        """Plot criteria for a given variable name.

        Args:
            variable: The name of the variable.
            title: A plot title.
            save: If True, save the plot on the disk.
            show: If True, show the plot.
            n_legend_cols: The number of text columns in the upper legend.
            directory: The directory path, either absolute or relative.

        Raises:
            ValueError: If the variable is missing from the dataset.
        """
        if variable not in self.names:
            raise ValueError(
                "The variable '{}' is missing from the dataset."
                "Available ones are: {}.".format(variable, ", ".join(self.names))
            )
        criteria, is_p_value = self.get_criteria(variable)
        x_values = []
        y_values = []
        labels = []
        x_value = 0
        for distribution, criterion in criteria.items():
            x_value += 1
            x_values.append(x_value)
            y_values.append(criterion)
            labels.append(distribution)
        plt.subplot(121)
        plt.bar(x_values, y_values, tick_label=labels, align="center")
        if is_p_value:
            plt.ylabel("p-value from {} test".format(self.fitting_criterion))
            plt.axhline(self.level, color="r", linewidth=2.0)
        plt.grid(True, "both")
        plt.subplot(122)
        data = array(self.dataset[variable])
        data_min = min(data)
        data_max = max(data)
        x_values = linspace(data_min, data_max, 1000)
        distributions = self._all_distributions[variable]
        try:
            plt.hist(data, density=True)
        except AttributeError:
            plt.hist(data, normed=True)
        for dist_name, dist_value in distributions.items():
            pdf = dist_value["fitted_distribution"].distribution.computePDF
            y_values = [pdf([x_value])[0] for x_value in x_values]
            plt.plot(x_values, y_values, label=dist_name, linewidth=2.0)
        plt.legend(
            bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
            loc="lower left",
            ncol=n_legend_cols,
            mode="expand",
            borderaxespad=0.0,
        )
        plt.grid(True, "both")
        if title is not None:
            plt.suptitle(title)
        filename = os.path.join(directory, "criteria.pdf")
        if save:
            plt.savefig(filename)
        if show:
            plt.show()
        plt.close()

    def _select_best_distributions(
        self, distributions_names  # type: Sequence[str]
    ):  # type: (...) -> Dict[str,Dict[str,Union[str,OTDistribution]]]
        """Select the best distributions for the different variables.

        Args:
            distributions_names: The names of the distributions.

        Returns:
            The best distributions for the different variables.
        """
        LOGGER.info("Select the best distribution for each variable.")
        distributions = {}
        for variable in self.names:
            all_distributions = self._all_distributions[variable]
            criteria = [
                all_distributions[distribution]["criterion"]
                for distribution in distributions_names
            ]
            select_from_measures = OTDistributionFitter.select_from_measures
            index = select_from_measures(
                criteria, self.fitting_criterion, self.level, self.selection_criterion
            )
            name = distributions_names[index]
            value = all_distributions[name]["fitted_distribution"]
            distributions[variable] = {"name": name, "value": value}
            LOGGER.info("| The best distribution for %s is %s.", variable, value)
        return distributions

    def _fit_distributions(
        self,
        distributions,  # type: Iterable[str]
    ):  # type: (...) -> Dict[str,Dict[str,Dict[str,Union[OTDistribution,MeasureType]]]]
        """Fit different distributions for the different marginals.

        Args:
            distributions: The distributions names.

        Returns:
            dict(str, dict): The distributions for the different variables.
        """
        dist_list = ", ".join(distributions)
        LOGGER.info(
            "Fit different distributions (%s) per variable "
            "and compute the goodness-of-fit criterion.",
            dist_list,
        )
        results = {}
        for variable in self.names:
            LOGGER.info("| Fit different distributions for %s.", variable)
            dataset = self.dataset[variable]
            results[variable] = self._fit_marginal_distributions(
                variable, dataset, distributions
            )
        return results

    def _fit_marginal_distributions(
        self,
        variable,  # type: str
        sample,  # type: ndarray
        distributions,  # type: Iterable[str]
    ):  # type: (...) -> Dict[str,Dict[str,Union[OTDistribution,MeasureType]]]
        """Fit different distributions for a given dataset marginal.

        Args:
            variable: A variable name.
            sample: A data array.
            distributions: The names of the distributions.

        Returns:
            The distributions for the different variables.
        """
        result = {}
        factory = OTDistributionFitter(variable, sample)
        for distribution in distributions:
            fitted_distribution = factory.fit(distribution)
            test_result = factory.compute_measure(
                fitted_distribution, self.fitting_criterion, self.level
            )
            result[distribution] = {}
            result[distribution]["fitted_distribution"] = fitted_distribution
            result[distribution]["criterion"] = test_result
        return result

    def compute_maximum(self):  # type: (...) -> Dict[str, ndarray]  # noqa: D102
        result = {
            name: self.distributions[name]["value"].math_upper_bound
            for name in self.names
        }
        return result

    def compute_mean(self):  # type: (...) -> Dict[str, ndarray]  # noqa: D102
        result = {name: self.distributions[name]["value"].mean for name in self.names}
        return result

    def compute_minimum(self):  # type: (...) -> Dict[str, ndarray] # noqa: D102
        result = {
            name: self.distributions[name]["value"].math_lower_bound
            for name in self.names
        }
        return result

    def compute_probability(
        self,
        thresh,  # type: float
        greater=True,  # type: bool
    ):  # type: (...) -> Dict[str, ndarray] # noqa: D102
        dist = self.distributions
        if greater:
            result = {
                name: 1 - dist[name]["value"].compute_cdf(thresh[name])[0]
                for name in self.names
            }
        else:
            result = {
                name: dist[name]["value"].compute_cdf(thresh[name])[0]
                for name in self.names
            }
        return result

    def compute_tolerance_interval(
        self,
        coverage,  # type: float
        confidence=0.95,  # type: float
        side=ToleranceIntervalSide.BOTH,  # type: ToleranceIntervalSide
    ):  # type: (...) -> Dict[str, Tuple[ndarray,ndarray]]
        # noqa: D102 D205 D212 D415
        if not 0.0 <= coverage <= 1.0:
            raise ValueError("The argument 'coverage' must be number in [0,1].")
        if not 0.0 <= confidence <= 1.0:
            raise ValueError("The argument 'confidence' must be number in [0,1].")
        limits = {}
        factory = ToleranceIntervalFactory()
        for variable in self.names:
            distribution = self.distributions[variable]
            cls = factory.get_class(distribution["name"])
            parameters = distribution["value"].marginals[0].getParameter()
            tolerance_interval = cls(self.n_samples, *parameters)
            limits[variable] = tolerance_interval.compute(coverage, confidence, side)
        return limits

    def compute_quantile(
        self,
        prob,  # type:float
    ):  # type: (...) -> Dict[str,ndarray] # noqa: D102
        prob = array([prob])
        result = {
            name: self.distributions[name]["value"].compute_inverse_cdf(prob)
            for name in self.names
        }
        return result

    def compute_standard_deviation(
        self,
    ):  # type: (...) -> Dict[str, ndarray]  # noqa: D102
        result = {
            name: self.distributions[name]["value"].standard_deviation
            for name in self.names
        }
        return result

    def compute_variance(self):  # type: (...) -> Dict[str, ndarray]  # noqa: D102
        result = {
            name: self.distributions[name]["value"].standard_deviation ** 2
            for name in self.names
        }
        return result

    def compute_moment(
        self,
        order,  # type: int
    ):  # type: (...) -> Dict[str, ndarray]  # noqa: D102
        dist = self.distributions
        result = [
            dist[name]["value"].distribution.getMoment(order)[0] for name in self.names
        ]
        return result

    def compute_range(self):  # type: (...) -> Dict[str, ndarray]  # noqa: D102
        result = {}
        for name in self.names:
            dist = self.distributions[name]["value"]
            result[name] = dist.math_upper_bound - dist.math_lower_bound
        return result



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                        documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
#        :author: Pierre-Jean Barjhoux, Benoit Pauwels - MDOScenarioAdapter
#                                                        Jacobian computation
"""
A Scenario which driver is an optimization algorithm
****************************************************
"""
from __future__ import division, unicode_literals

import logging
from copy import copy, deepcopy
from datetime import timedelta
from timeit import default_timer as timer

from numpy import atleast_1d, zeros
from numpy.linalg import norm

from gemseo.algos.lagrange_multipliers import LagrangeMultipliers
from gemseo.algos.opt.opt_factory import OptimizersFactory
from gemseo.algos.post_optimal_analysis import PostOptimalAnalysis
from gemseo.core.discipline import MDODiscipline
from gemseo.core.json_grammar import JSONGrammar
from gemseo.core.parallel_execution import DiscParallelLinearization
from gemseo.core.scenario import Scenario

# The detection of formulations requires to import them,
# before calling get_formulation_from_name


LOGGER = logging.getLogger(__name__)


class MDOScenario(Scenario):
    """Multidisciplinary Design Optimization Scenario, main user interface Creates an
    optimization problem and solves it with an optimizer.

    The main differences between Scenario and MDOScenario are the allowed
    inputs in the MDOScenario.json, which differs from DOEScenario.json,
    at least on the driver names

    MDO Problem description: links the disciplines and the formulation
    to create an optimization problem.
    Use the class by instantiation.

    Create your disciplines beforehand.

    Specify the formulation by giving the class name such as the string
    "MDF"

    The reference_input_data is the typical input data dict that is provided
    to the run method of the disciplines

    Specify the objective function name, which must be an output
    of a discipline of the scenario, with the "objective_name" attribute

    If you want to add additional design constraints,
    use the add_user_defined_constraint method

    To view the results, use the "post_process" method after execution.
    You can view:

    - the design variables history, the objective value, the constraints,
      by using:
      scenario.post_process("OptHistoryView", show=False, save=True)
    - Quadratic approximations of the functions close to the
      optimum, when using gradient based algorithms, by using:
      scenario.post_process("QuadApprox", method="SR1", show=False,
      save=True, function="my_objective_name",
      file_path="appl_dir")
    - Self Organizing Maps of the design space, by using:
      scenario.post_process("SOM", save=True, file_path="appl_dir")

    To list post-processing on your setup,
    use the method scenario.posts
    For more detains on their options, go to the "gemseo.post" package
    """

    # Constants for input variables in json schema
    MAX_ITER = "max_iter"
    X_OPT = "x_opt"

    def __init__(
        self,
        disciplines,
        formulation,
        objective_name,
        design_space,
        name=None,
        **formulation_options
    ):
        """Constructor, initializes the MDO scenario Objects instantiation and checks
        are made before run intentionally.

        :param disciplines: the disciplines of the scenario
        :param formulation: the formulation name,
            the class name of the formulation in gemseo.formulations
        :param objective_name: the objective function name
        :param design_space: the design space
        :param name: scenario name
        :param formulation_options: options for creation of the formulation
        """
        # This loads the right json grammars from class name
        super(MDOScenario, self).__init__(
            disciplines,
            formulation,
            objective_name,
            design_space,
            name,
            **formulation_options
        )
        self.clear_history_before_run = False

    def _run_algorithm(self):
        """Runs the optimization algo."""
        problem = self.formulation.opt_problem
        # Clears the database when multiple runs are performed (bi level)
        if self.clear_history_before_run:
            problem.database.clear()
        algo_name = self.local_data[self.ALGO]
        max_iter = self.local_data[self.MAX_ITER]
        options = self.local_data.get(self.ALGO_OPTIONS)
        if options is None:
            options = {}
        if self.MAX_ITER in options:
            LOGGER.warning(
                "Double definition of algorithm option " "max_iter, keeping value: %s",
                max_iter,
            )
            options.pop(self.MAX_ITER)

        lib = self._algo_factory.create(algo_name)
        self.optimization_result = lib.execute(
            problem, algo_name=algo_name, max_iter=max_iter, **options
        )
        return self.optimization_result

    def _run(self):
        """Execute the scenario and run the optimization problems."""
        t_0 = timer()
        LOGGER.info(" ")
        LOGGER.info("*** Start MDO Scenario execution ***")
        LOGGER.info("%s", repr(self))
        self._run_algorithm()
        # MDODiscipline.execute is not finished therefore self.exec_time is not
        # computed yet, need to recompute it, besides exec_time is the total
        # execution time, while this is for a single execution
        delta_t = timer() - t_0
        LOGGER.info(
            "*** MDO Scenario run terminated in %s ***", timedelta(seconds=delta_t)
        )

    def _init_algo_factory(self):
        """Initalizes the algorithms factory."""
        self._algo_factory = OptimizersFactory()


class MDOScenarioAdapter(MDODiscipline):
    """An adapter class for MDO Scenario:

    input variables are specified they update default_data in the top level discipline
    they output data from the top level discipline outputs.
    """

    LOWER_BND_SUFFIX = "_lower_bnd"
    UPPER_BND_SUFFIX = "_upper_bnd"
    MULTIPLIER_SUFFIX = "_multiplier"

    def __init__(
        self,
        scenario,
        inputs_list,
        outputs_list,
        reset_x0_before_opt=False,
        set_x0_before_opt=False,
        set_bounds_before_opt=False,
        cache_type=MDODiscipline.SIMPLE_CACHE,
        output_multipliers=False,
    ):
        """Initialize the scenario adapter.

        :param scenario: the scenario to adapt
        :type scenario: MDOScenario
        :param inputs_list: list of inputs to overload at
            sub scenario execution
        :type inputs_list: list(str)
        :param outputs_list: list of outputs to get from
            scenario execution
        :type outputs_list: list(str)
        :param reset_x0_before_opt: before running the sub optimization, reset
            the initial guess
        :type reset_x0_before_opt: bool
        :param set_x0_before_opt: if True, sets the initial point of the sub
            scenario, useful for multi-start
        :type set_x0_before_opt: bool
        :param set_bounds_before_opt: if True, sets the bounds of the design
            space, useful for trust regions
        :type set_bounds_before_opt: bool
        :param cache_type: type of cache policy, SIMPLE_CACHE
            or HDF5_CACHE
        :type cache_type: str
        :param output_multipliers: if True then the Lagrange multipliers of the
            scenario optimal solution are computed and added to the outputs
        :type output_multipliers: bool
        """
        if reset_x0_before_opt and set_x0_before_opt:
            raise ValueError("Inconsistent options for ScenarioAdapter !")
        self.scenario = scenario
        self._set_x0_before_opt = set_x0_before_opt
        self._set_bounds_before_opt = set_bounds_before_opt
        self._inputs_list = inputs_list
        self._outputs_list = outputs_list
        self._reset_x0_before_opt = reset_x0_before_opt
        self._output_multipliers = output_multipliers
        name = "{}_adapter".format(scenario.name)
        super(MDOScenarioAdapter, self).__init__(name, cache_type=cache_type)

        self._update_grammars()
        self._dv_in_names = None
        if set_x0_before_opt:
            dv_names = set(self.scenario.formulation.design_space.variables_names)
            self._dv_in_names = list(dv_names & set(self._inputs_list))

        # Set the initial bounds as default bounds
        self._bounds_names = []
        if set_bounds_before_opt:
            dspace = scenario.design_space
            lower_bounds = dspace.array_to_dict(dspace.get_lower_bounds())
            lower_suffix = MDOScenarioAdapter.LOWER_BND_SUFFIX
            upper_bounds = dspace.array_to_dict(dspace.get_upper_bounds())
            upper_suffix = MDOScenarioAdapter.UPPER_BND_SUFFIX
            for bounds, suffix in [
                (lower_bounds, lower_suffix),
                (upper_bounds, upper_suffix),
            ]:
                bounds = {name + suffix: val for name, val in bounds.items()}
                self.default_inputs.update(bounds)
                self._bounds_names.extend(bounds.keys())

        # Optimization functions are redefined at each run
        # since default inputs of top
        # level discipline change
        # History must be erased otherwise the wrong values are retrieved
        # between two runs
        scenario.clear_history_before_run = True
        self._x_dict_0 = deepcopy(scenario.design_space.get_current_x_dict())

        self.post_optimal_analysis = None

    def _update_grammars(self):
        """Update the input and output grammars."""
        formulation = self.scenario.formulation
        opt_problem = formulation.opt_problem
        top_leveld = formulation.get_top_level_disc()
        for disc in top_leveld:
            self.input_grammar.update_from(disc.input_grammar)
            self.output_grammar.update_from(disc.output_grammar)
            # The output may also be the optimum value of the design
            # variables, so the output grammar may contain inputs
            # of the disciplines. All grammars are filtered just after
            # this loop
            self.output_grammar.update_from(disc.input_grammar)
            self.default_inputs.update(disc.default_inputs)

        self.input_grammar.restrict_to(self._inputs_list)
        self.output_grammar.restrict_to(self._outputs_list)
        # If a DV is not an input of the top level disciplines:
        output_names = self.output_grammar.get_data_names()
        missing_out = set(self._outputs_list) - set(output_names)
        if missing_out:
            dv_names = opt_problem.design_space.variables_names
            miss_dvs = set(dv_names) & set(missing_out)
            if miss_dvs:
                dv_gram = JSONGrammar("dvs")
                dv_gram.initialize_from_data_names(miss_dvs)
                self.output_grammar.update_from(dv_gram)

        output_names = self.output_grammar.get_data_names()
        missing_out = set(self._outputs_list) - set(output_names)

        # Add the design variables bounds to the input grammar
        if self._set_bounds_before_opt:
            current_x = self.scenario.design_space.get_current_x_dict()
            typical_data_dict = dict()
            for suffix in [
                MDOScenarioAdapter.LOWER_BND_SUFFIX,
                MDOScenarioAdapter.UPPER_BND_SUFFIX,
            ]:
                bnds = {name + suffix: val for name, val in current_x.items()}
                typical_data_dict.update(bnds)
            bnds_gram = JSONGrammar("bnds")
            bnds_gram.initialize_from_base_dict(typical_data_dict)
            self.input_grammar.update_from(bnds_gram)

        if missing_out:
            raise ValueError(
                "Can't compute outputs from scenarios: " + str(missing_out)
            )
        missing_inpt = set(self._inputs_list) - set(self.input_grammar.get_data_names())
        if missing_inpt:
            raise ValueError(
                "Can't compute inputs from scenarios: " + str(missing_inpt)
            )

        # Add the Lagrange multipliers to the output grammar
        if self._output_multipliers:
            self._add_output_multipliers()

    def _add_output_multipliers(self):
        """Add the Lagrange multipliers of the scenario optimal solution as outputs of
        the adapter."""
        # Fill a dictionary with data of typical shapes
        base_dict = dict()
        problem = self.scenario.formulation.opt_problem
        # bound-constraints multipliers
        current_x = problem.design_space.get_current_x_dict()
        base_dict.update(
            {
                self.get_bnd_mult_name(var_name, False): val
                for var_name, val in current_x.items()
            }
        )
        base_dict.update(
            {
                self.get_bnd_mult_name(var_name, True): val
                for var_name, val in current_x.items()
            }
        )
        # equality- and inequality-constraints multipliers
        base_dict.update(
            {
                self.get_cstr_mult_name(cstr_name): zeros(1)
                for cstr_name in problem.get_constraints_names()
            }
        )

        # Update the output grammar
        multipliers_grammar = JSONGrammar("multipliers")
        multipliers_grammar.initialize_from_base_dict(base_dict)
        self.output_grammar.update_from(multipliers_grammar)

    @staticmethod
    def get_bnd_mult_name(variable_name, is_upper):
        """Return the name of the lower bound-constraint multiplier of a variable.

        :param variable_name: name of the variable
        :type variable_name: str
        :param is_upper: if True then return the upper bound-constraint multiplier
            name, otherwise return the lower bound-constraint multiplier
        :type is_upper: bool
        :return: name of the lower bound-constraint multiplier
        :rtype: str
        """
        mult_name = variable_name
        mult_name += "_upp-bnd" if is_upper else "_low-bnd"
        mult_name += MDOScenarioAdapter.MULTIPLIER_SUFFIX
        return mult_name

    @staticmethod
    def get_cstr_mult_name(constraint_name):
        """Return the name of the multiplier of a constraint.

        :param constraint_name: name of the constraint
        :type constraint_name: str
        :return: name of the multiplier
        :rtype: str
        """
        return constraint_name + MDOScenarioAdapter.MULTIPLIER_SUFFIX

    def _run(self):
        """Runs the scenario."""
        self._pre_run()
        self.scenario.execute()
        self._post_run()

    def _pre_run(self):
        """Pre-processes the scenario."""
        formulation = self.scenario.formulation
        opt_problem = formulation.opt_problem
        design_space = opt_problem.design_space
        top_leveld = formulation.get_top_level_disc()

        # Update the top level discipline default inputs with adapter inputs
        # This is the key role of the adapter
        for indata in self._inputs_list:
            for disc in top_leveld:
                if disc.is_input_existing(indata):
                    disc.default_inputs[indata] = self.local_data[indata]

        # Default inputs have changed, therefore caches shall be cleared
        self.scenario.cache.clear()
        self.scenario.reset_statuses_for_run()
        for func in opt_problem.get_all_functions():
            # Avoids max_iter reached
            func.n_calls = 0

        if self._reset_x0_before_opt:
            design_space.set_current_x(self._x_dict_0)

        # Set the starting point of the sub scenario with current dv names
        if self._set_x0_before_opt:
            dv_values = {dv_n: self.local_data[dv_n] for dv_n in self._dv_in_names}
            self.scenario.formulation.design_space.set_current_x(dv_values)

        # Set the bounds of the sub-scenario
        if self._set_bounds_before_opt:
            for name in design_space.variables_names:
                # Set the lower bound
                lower_suffix = MDOScenarioAdapter.LOWER_BND_SUFFIX
                lower_bound = self.local_data[name + lower_suffix]
                design_space.set_lower_bound(name, lower_bound)
                # Set the upper bound
                upper_suffix = MDOScenarioAdapter.UPPER_BND_SUFFIX
                upper_bound = self.local_data[name + upper_suffix]
                design_space.set_upper_bound(name, upper_bound)

    def _post_run(self):
        """Post-process the scenario."""
        formulation = self.scenario.formulation
        opt_problem = formulation.opt_problem
        design_space = opt_problem.design_space

        # Test if the last evaluation is the optimum
        x_opt = design_space.get_current_x()
        last_x = opt_problem.database.get_x_by_iter(-1)
        last_eval_not_opt = norm(x_opt - last_x) / (1.0 + norm(last_x)) > 1e-14
        if last_eval_not_opt:
            # Revaluate all functions at optimum
            # To re execute all disciplines and get the right data
            opt_problem.evaluate_functions(
                x_opt,
                eval_jac=False,
                eval_obj=True,
                normalize=False,
                # Force call without database
                no_db_no_norm=True,
            )

        # Retrieves top-level discipline outputs
        self._retrieve_top_level_outputs()

        # Compute the Lagrange multipliers and store them in the local data
        if self._output_multipliers:
            self._compute_lagrange_multipliers()

    def _retrieve_top_level_outputs(self):
        """Overwrites the adapter outputs with the top-level discipline outputs and
        optimal design parameters."""
        formulation = self.scenario.formulation
        opt_problem = formulation.opt_problem
        top_leveld = formulation.get_top_level_disc()
        x_dict = opt_problem.design_space.get_current_x_dict()
        for outdata in self._outputs_list:
            for disc in top_leveld:
                if disc.is_output_existing(outdata) and outdata not in x_dict:
                    self.local_data[outdata] = disc.local_data[outdata]
            out_ds = x_dict.get(outdata)
            if out_ds is not None:
                self.local_data[outdata] = out_ds

    def _compute_lagrange_multipliers(self):
        """Compute the Lagrange multipliers for the optimal solution of the scenario and
        store them in the local data."""
        # Compute the Lagrange multipliers
        problem = self.scenario.formulation.opt_problem
        x_opt = problem.solution.x_opt
        lagrange = LagrangeMultipliers(problem)
        lagrange.compute(x_opt, problem.ineq_tolerance)

        # Store the Lagrange multipliers in the local data
        multipliers = lagrange.get_multipliers_arrays()
        self.local_data.update(
            {
                self.get_bnd_mult_name(name, False): mult
                for name, mult in multipliers[lagrange.LOWER_BOUNDS].items()
            }
        )
        self.local_data.update(
            {
                self.get_bnd_mult_name(name, True): mult
                for name, mult in multipliers[lagrange.UPPER_BOUNDS].items()
            }
        )
        self.local_data.update(
            {
                self.get_cstr_mult_name(name): mult
                for name, mult in multipliers[lagrange.EQUALITY].items()
            }
        )
        self.local_data.update(
            {
                self.get_cstr_mult_name(name): mult
                for name, mult in multipliers[lagrange.INEQUALITY].items()
            }
        )

    def get_expected_workflow(self):
        return self.scenario.get_expected_workflow()

    def get_expected_dataflow(self):
        return self.scenario.get_expected_dataflow()

    def _compute_jacobian(self, inputs=None, outputs=None):
        """
        Computes the Jacobian of the adapted scenario outputs with respect to
        its inputs.
        The Jacobian is stored as a dict of ndarray dict:
        jac = {name: { input_name: (output_dim, input_dim) ndarray } }

        The bound-constraints on the scenario optimization variables are
        assumed independent of the other scenario inputs.

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization should
            be performed wrt all inputs (Default value = None)
        :param outputs: linearization should be performed on outputs list.
            If None, linearization should be performed
            on all outputs (Default value = None)
        """
        opt_problem = self.scenario.formulation.opt_problem
        ineq_tol = opt_problem.ineq_tolerance
        outvars = opt_problem.objective.outvars
        if len(outvars) != 1:
            raise ValueError("The objective must be single-valued.")

        # Check the required inputs
        if inputs is None and not self._set_bounds_before_opt:
            inputs = self._inputs_list
        elif inputs is None and self._set_bounds_before_opt:
            # Bounds are inputs of the adapter
            inputs = [
                name for name in self._bounds_names if name not in self._inputs_list
            ]
            inputs = self._inputs_list + inputs
        elif set(inputs) - set(self._inputs_list) - set(self._bounds_names):
            not_inputs = set(inputs) - set(self._inputs_list) - set(self._bounds_names)
            raise ValueError(
                "The following are not inputs of the adapter: "
                + ", ".join(not_inputs)
                + "."
            )
        # N.B the adapter is assumed constant w.r.t. bounds
        bound_inputs = set(inputs) & set(self._bounds_names)

        # Check the required outputs
        if outputs is None:
            outputs = outvars
        elif set(outputs) - set(self._outputs_list):
            raise ValueError(
                "The following are not outputs of the adapter: "
                + str(set(outputs) - set(self._outputs_list))
                + "."
            )
        nondifferentiable_outputs = set(outputs) - set(outvars)
        if nondifferentiable_outputs:
            raise ValueError(
                "Post-optimal Jacobians of "
                + ", ".join(nondifferentiable_outputs)
                + " cannot be computed."
            )

        # Initialize the Jacobian
        diff_inputs = [name for name in inputs if name not in bound_inputs]
        # N.B. there may be only bound inputs
        self._init_jacobian(diff_inputs, outputs)

        # Compute the Jacobians of the optimization functions
        jacobians = self._compute_auxiliary_jacobians(diff_inputs, use_threading=True)

        # Perform the post-optimal analysis
        self.post_optimal_analysis = PostOptimalAnalysis(opt_problem, ineq_tol)
        post_opt_jac = self.post_optimal_analysis.execute(
            outputs, diff_inputs, jacobians
        )
        self.jac.update(post_opt_jac)

        # Fill the Jacobian blocks w.r.t. bounds with zeros
        for out_jac in self.jac.values():
            for in_name in bound_inputs:
                in_dim = self.default_inputs[in_name].size
                out_jac[in_name] = zeros((1, in_dim))

    def _compute_auxiliary_jacobians(self, inputs, func_names=None, use_threading=True):
        """Computes the Jacobians of the optimization functions.

        :param inputs: names list of the inputs w.r.t. which differentiate
        :type inputs: list(str)
        :param func_names: names list of the functions to differentiate
            If None then all the optimizations functions are differentiated
        :type func_names: list(str)
        :param use_threading : if True, use Threads instead of processes
            to parallelize the execution
        :type use_threading: bool
        """
        # Gather the names of the functions to differentiate
        opt_problem = self.scenario.formulation.opt_problem
        if func_names is None:
            func_names = (
                opt_problem.objective.outvars + opt_problem.get_constraints_names()
            )

        # Identify the disciplines that compute the functions
        disc_dict = dict()
        for name in func_names:
            for disc in self.scenario.formulation.get_top_level_disc():
                if disc.is_all_outputs_existing([name]):
                    disc_dict[name] = disc
                    break

        # Linearize the required disciplines
        for disc in set(disc_dict.values()):
            inputs_set = set(disc.get_input_data_names()) & set(inputs)
            outputs_set = set(disc.get_output_data_names()) & set(func_names)
            if inputs_set and outputs_set:
                disc.add_differentiated_inputs(list(inputs_set))
                disc.add_differentiated_outputs(list(outputs_set))
        disc_list = list(set(disc_dict.values()))
        paralell_lin = DiscParallelLinearization(disc_list, use_threading=use_threading)
        # Update the local data with the optimal design parameters
        # [The adapted scenario is assumed to have been run beforehand.]
        x_opt_dict = opt_problem.design_space.get_current_x_dict()
        post_opt_data = copy(self.local_data)
        post_opt_data.update(x_opt_dict)
        paralell_lin.execute([post_opt_data] * len(disc_list))

        # Store the Jacobians
        jacobians = dict()
        for name in func_names:
            jacobians[name] = dict()
            for input_name in inputs:
                jac_block = disc_dict[name].jac[name].get(input_name)
                if jac_block is None:
                    output_value = self.get_outputs_by_name(name)
                    input_value = self.get_inputs_by_name(input_name)
                    jac_block = zeros((len(output_value, len(input_value))))
                jacobians[name][input_name] = jac_block

        return jacobians

    def add_outputs(self, outputs_names):
        """Add outputs to the scenario adapter.

        :param outputs_names: names of the outputs to be added
        :type outputs_names: list(str)
        """
        names_to_add = [
            name for name in outputs_names if name not in self._outputs_list
        ]
        self._outputs_list.extend(names_to_add)
        self._update_grammars()


class MDOObjScenarioAdapter(MDOScenarioAdapter):
    """A scenario adapter that overwrites the local data with the optimal objective
    function value."""

    def _retrieve_top_level_outputs(self):
        """Overwrites the adapter outputs with the top-level discipline outputs and
        optimal design parameters."""
        formulation = self.scenario.formulation
        opt_problem = formulation.opt_problem
        top_leveld = formulation.get_top_level_disc()

        # Get the optimal outputs
        optim_data = opt_problem.design_space.get_current_x_dict()
        f_opt = opt_problem.get_optimum()[0]
        if not opt_problem.minimize_objective:
            f_opt = -f_opt
        outvars = opt_problem.objective.outvars
        if not len(outvars) == 1:
            raise ValueError("The objective function must be single-valued.")
        optim_data[outvars[0]] = atleast_1d(f_opt)

        # Overwrite the adapter local data
        for outdata in self._outputs_list:
            for disc in top_leveld:
                if disc.is_output_existing(outdata) and outdata not in optim_data:
                    self.local_data[outdata] = disc.local_data[outdata]
            out_ds = optim_data.get(outdata)
            if out_ds is not None:
                self.local_data[outdata] = out_ds

    def _compute_jacobian(self, inputs=None, outputs=None):
        """
        Computes the Jacobian of the adapted scenario outputs with respect to
        its inputs.
        The Jacobian is stored as a dict of ndarray dict:
        jac = {name: { input_name: (output_dim, input_dim) ndarray } }

        The bound-constraints on the scenario optimization variables are
        assumed independent of the other scenario inputs.

        :param inputs: linearization should be performed with respect
            to inputs list. If None, linearization should
            be performed wrt all inputs (Default value = None)
        :param outputs: linearization should be performed on outputs list.
            If None, linearization should be performed
            on all outputs (Default value = None)
        """
        MDOScenarioAdapter._compute_jacobian(self, inputs, outputs)
        # The gradient of the objective function cannot be computed by the
        # disciplines, but the gradients of the constraints can.
        # The objective function is assumed independent of non-optimization
        # variables.
        obj_name = self.scenario.formulation.opt_problem.objective.outvars[0]
        mult_cstr_jac_key = PostOptimalAnalysis.MULT_DOT_CONSTR_JAC
        self.jac[obj_name] = dict(self.jac[mult_cstr_jac_key])



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Draw lines from a :class:`.Dataset`.

A :class:`.Lines` plot represents variables vs samples using lines.
"""
from __future__ import division, unicode_literals

from typing import List, Mapping, Optional, Sequence

import matplotlib.pyplot as plt
from matplotlib.figure import Figure

from gemseo.post.dataset.dataset_plot import DatasetPlot, DatasetPlotPropertyType


class Lines(DatasetPlot):
    """Plot sampled variables as lines."""

    def _plot(
        self,
        properties,  # type: Mapping[str,DatasetPlotPropertyType]
        variables=None,  # type: Optional[Sequence[str]]
    ):  # type: (...) -> List[Figure]
        """
        Args:
            variables: The names of the variables to plot.
        """
        x_data = range(len(self.dataset))
        if variables is None:
            y_data = self.dataset.get_all_data(False, True)
            variables = y_data.keys()
        else:
            y_data = self.dataset[variables]

        plt.figure(figsize=self.figsize)
        self._set_color(properties, len(variables))
        self._set_linestyle(properties, len(variables), "-")
        index = 0
        for name, value in y_data.items():
            plt.plot(
                x_data,
                value,
                linestyle=self.linestyle[index],
                color=self.color[index],
                label=name,
            )
            index += 1
        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)
        plt.title(self.title)
        plt.legend(loc=self.legend_location)
        return [plt.gcf()]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Constraints aggregation core functions."""

from __future__ import division, unicode_literals

from math import log
from typing import Optional, Sequence, Union

from numpy import argmax as np_argmax
from numpy import array, atleast_2d
from numpy import exp as np_exp
from numpy import max as np_max
from numpy import multiply, ndarray, ones
from numpy import sum as np_sum
from numpy import zeros


def ks_agg(
    orig_val,  # type: ndarray
    indices=None,  # type: Optional[Sequence[int]]
    rho=1e2,  # type: float
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> float
    """Transform a vector of equalities into a Kreisselmeier–Steinhauser function.

    Kreisselmeier G, Steinhauser R (1983)
    Application of Vector Performance Optimization
    to a Robust Control Loop Design for a Fighter Aircraft.
    International Journal of Control 37(2):251–284,
    doi:10.1080/00207179.1983.9753066

    Graeme J. Kennedy, Jason E. Hicken,
    Improved constraint-aggregation methods,
    Computer Methods in Applied Mechanics and Engineering,
    Volume 289,
    2015,
    Pages 332-354,
    ISSN 0045-7825,
    https://doi.org/10.1016/j.cma.2015.02.017.
    (http://www.sciencedirect.com/science/article/pii/S0045782515000663)

    Args:
        orig_val: The original constraint values.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        rho: The multiplicative parameter in the exponential.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated function value.
    """
    if indices is not None:
        orig_val = orig_val[indices]

    orig_val *= scale
    alpha = len(orig_val)
    m = max(orig_val)

    return (
        m
        + (1.0 / rho) * log((1.0 / alpha) * sum(np_exp(rho * (orig_val + 1.0 - m))))
        - 1.0
    )


def ks_agg_jac_v(
    orig_val,  # type: ndarray
    orig_jac,  # type: ndarray
    indices=None,  # type: Optional[Sequence[int]]
    rho=1e2,  # type: float
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> ndarray
    """Aggregate inequality constraints.

    Jacobian vector product of the constraints aggregation method for inequality
    constraints.

    See :cite:t:`kennedy2015improved` and  :cite:t:`kreisselmeier1983application`.

    Args:
        orig_val: The original constraint values.
        orig_jac: The original constraint jacobian.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        rho: The multiplicative parameter in the exponential.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated function Jacobian vector product.
    """
    if indices is not None:
        orig_jac = orig_jac[indices, :]
        orig_val = orig_val[indices]

    orig_jac *= scale
    orig_val *= scale

    m = max(orig_val)
    div = np_sum(np_exp(rho * (orig_val + 1.0 - m)))
    weights = np_exp(rho * (orig_val + 1.0 - m)).T / div
    return np_sum(multiply(atleast_2d(weights).T, orig_jac), axis=0)


def ks_agg_jac(
    orig_val,  # type: ndarray
    indices=None,  # type: Optional[Sequence[int]]
    rho=1e2,  # type: float
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> ndarray
    """Transform a vector of equalities into a scalar equivalent constraint.

    Jacobian of the Constraints aggregation method for inequality constraints.

    See :cite:t:`kennedy2015improved` and  :cite:t:`kreisselmeier1983application`.

    Args:
        orig_val: The original constraint values.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        rho: The multiplicative parameter in the exponential.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated function Jacobian.
    """
    full_size = orig_val.size
    if indices is not None:
        orig_val = orig_val[indices]

    orig_val *= scale

    m = max(orig_val)
    div = np_sum(np_exp(rho * (orig_val + 1.0 - m)))
    weights = np_exp(rho * (orig_val + 1.0 - m)).T / div
    der = atleast_2d(multiply(weights, scale))
    return __filter_jac(der, full_size, indices)


def iks_agg(
    orig_val,  # type: ndarray
    indices=None,  # type: Optional[Sequence[int]]
    rho=1e2,  # type: float
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> float
    """Aggregate IKS Constraints for inequality constraints.

    See :cite:t:`kennedy2015improved`.

    Args:
        orig_val: The original constraint values.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        rho: The multiplicative parameter in the exponential.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated value.
    """
    if indices is not None:
        orig_val = orig_val[indices]

    orig_val *= scale

    m = max(orig_val)
    iks = sum(orig_val * np_exp(rho * (orig_val + 1.0 - m)))
    iks /= sum(np_exp(rho * (orig_val + 1.0 - m)))

    return iks


def iks_agg_jac_v(
    orig_val,  # type: ndarray
    orig_jac,  # type: ndarray
    indices=None,  # type: Optional[Sequence[int]]
    rho=1e2,  # type: float
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> ndarray
    """Aggregate inequality constraints.

    Jacobian vector product of the IKS Constraints aggregation method for inequality
    constraints.

    See :cite:t:`kennedy2015improved`.

    Args:
        orig_val: The original constraint values.
        orig_jac: The original constraint jacobian.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        rho: The multiplicative parameter in the exponential.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated function Jacobian vector product.
    """
    if indices is not None:
        orig_jac = orig_jac[indices, :]
        orig_val = orig_val[indices]

    orig_jac *= scale
    orig_val *= scale

    m = max(orig_val)

    iks_num = sum(orig_val * np_exp(rho * (orig_val + 1.0 - m)))
    iks_den = sum(np_exp(rho * (orig_val + 1.0 - m)))

    iks_num_der = np_sum(
        multiply(atleast_2d(np_exp(rho * (orig_val + 1.0 - m))).T, orig_jac), axis=0
    )
    iks_num_der += np_sum(
        multiply(
            atleast_2d(np_exp(rho * (orig_val + 1.0 - m)) * orig_val).T,
            rho * orig_jac,
        ),
        axis=0,
    )
    iks_den_der = np_sum(
        multiply(atleast_2d(np_exp(rho * (orig_val + 1.0 - m))).T, rho * orig_jac),
        axis=0,
    )
    return (-iks_den_der / iks_den ** 2) * iks_num + iks_num_der / iks_den


def iks_agg_jac(
    orig_val,  # type: ndarray
    indices=None,  # type: Optional[Sequence[int]]
    rho=1e2,  # type: float
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> ndarray
    """Jacobian of the IKS Constraints aggregation method for inequality constraints.

    Kennedy, Graeme J., and Jason E. Hicken.
    "Improved constraint-aggregation methods."
    Computer Methods in Applied Mechanics and Engineering
    289 (2015): 332-354.

    Args:
        orig_val: The original constraint values.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        rho: The multiplicative parameter in the exponential.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated function Jacobian.
    """
    full_size = orig_val.size
    if indices is not None:
        orig_val = orig_val[indices]

    orig_val *= scale

    m = max(orig_val)

    iks_num = sum(orig_val * np_exp(rho * (orig_val + 1.0 - m)))
    iks_den = sum(np_exp(rho * (orig_val + 1.0 - m)))

    iks_num_der = atleast_2d(np_exp(rho * (orig_val + 1.0 - m)))
    iks_num_der += rho * atleast_2d(np_exp(rho * (orig_val + 1.0 - m)) * orig_val)
    iks_den_der = rho * atleast_2d(np_exp(rho * (orig_val + 1.0 - m)))
    iks_d = atleast_2d((-iks_den_der / iks_den ** 2) * iks_num + iks_num_der / iks_den)

    iks_d = multiply(iks_d, scale)

    return __filter_jac(iks_d, full_size, indices)


def __filter_jac(
    orig_jac,  # type: ndarray
    full_size,  # type: int
    indices,  # type: Sequence[int]
):  # type: (...) -> ndarray
    """Filters the Jacobian according to the indices.

    Args:
        orig_jac: The original Jacobian.
        full_size: The size of the full flatten Jacobian array.
        indices: The array of indices to filter.

    Returns:
        The filtered Jacobian.
    """
    if indices is not None:
        jac = zeros((1, full_size))
        jac[:, indices] = orig_jac
        return jac

    return orig_jac


def sum_square_agg(
    orig_val,  # type: ndarray
    indices=None,  # type: Optional[Sequence[int]]
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> ndarray
    """Transform a vector of equalities into a sum of squared constraints.

    Args:
        orig_val: The original constraint values.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, scale all the constraint values.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated function.
    """
    if indices is not None:
        orig_val = orig_val[indices]
    orig_val *= scale
    return np_sum(orig_val ** 2)


def sum_square_agg_jac_v(
    orig_val,  # type: ndarray
    orig_jac,  # type: ndarray
    indices=None,  # type: Optional[Sequence[int]]
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> ndarray
    """Transform a vector of equalities into a sum of squared constraints.

    Jacobian vector product of the constraints aggregation method for equality
    constraints.

    Args:
        orig_val: The original constraint values.
        orig_jac: The original constraint jacobian.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated function Jacobian vector product.
    """
    if indices is not None:
        orig_jac = orig_jac[indices, :]
        orig_val = orig_val[indices]
    orig_jac *= scale
    orig_val *= scale
    return 2 * np_sum(orig_jac * orig_val, axis=0)


def sum_square_agg_jac(
    orig_val,  # type: ndarray
    indices=None,  # type: Optional[Sequence[int]]
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> ndarray
    """Transform a vector of equalities into a sum of squared constraints.

    Jacobian of the constraints aggregation method for equality constraints.

    Args:
        orig_val: The original constraint values.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated function Jacobian.
    """
    if indices is not None:
        jac = zeros((1, orig_val.size))
        jac[:, indices] = 2.0
    else:
        jac = 2 * ones((1, orig_val.size))

    return jac


def max_agg(
    orig_val,  # type: ndarray
    indices=None,  # type: Optional[Sequence[int]]
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> float
    """Transform a vector of equalities into a max of all values.

    Constraints aggregation method for inequality constraints.

    Args:
        orig_val: The original constraint values.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated function.
    """
    if indices is not None:
        orig_val = orig_val[indices]
    orig_val *= scale
    return array([np_max(orig_val)])


def max_agg_jac_v(
    orig_val,  # type: ndarray
    orig_jac,  # type: ndarray
    indices=None,  # type: Optional[Sequence[int]]
    scale=1.0,  # type: Union[float, ndarray]
):  # type: (...) -> ndarray
    """Transform a vector of equalities into the max of all the values.

    Jacobian vector product of the max constraints aggregation method for inequality
    constraints.

    Args:
        orig_val: The original constraint values.
        orig_jac: The original constraint jacobian.
        indices: The indices to generate a subset of the outputs to aggregate.
            If ``None``, aggregate all the outputs.
        scale: The scaling factor for multiplying the constraints.

    Returns:
        The aggregated function.
    """
    if indices is not None:
        orig_jac = orig_jac[indices, :]
        orig_val = orig_val[indices]
    orig_jac *= scale
    orig_val *= scale
    i_max = np_argmax(orig_val)
    return atleast_2d(orig_jac)[i_max, :]



# -*- coding: utf-8 -*-
class JsonSchemaException(ValueError):
    """
    Exception raised by validation function. Contains ``message`` with
    information what is wrong.
    """

    def __init__(self, message):
        super(JsonSchemaException, self).__init__(message)
        self.message = message


class JsonSchemaDefinitionException(JsonSchemaException):
    """
    Exception raised by generator of validation function.
    """



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""An MDODiscipline to aggregates constraints using KS/IKS/Max methods."""

from typing import Any, Optional, Sequence

from numpy import atleast_1d

from gemseo.algos.aggregation.core import iks_agg, iks_agg_jac, ks_agg, ks_agg_jac
from gemseo.core.discipline import MDODiscipline
from gemseo.utils.data_conversion import DataConversion

METHODS_MAP = {"KS": ks_agg, "IKS": iks_agg}
METHODS_JAC_MAP = {"KS": ks_agg_jac, "IKS": iks_agg_jac}


class ConstrAggegationDisc(MDODiscipline):
    """A discipline that aggregates the constraints computed by other disciplines.

    An efficient alternative to constraint aggregation in the optimization problem
    is to aggregate the constraint in a discipline.

    This can be included in a MDO formulation,
    and in particular in a MDA,
    so only one adjoint calculation can be performed for the aggregated
    constraint instead of one adjoint per original constraint dimension.

    See :cite:t:`kennedy2015improved` and  :cite:t:`kreisselmeier1983application`.
    """

    def __init__(
        self,
        constr_data_names,  # type: Sequence[str]
        method_name,  # type: str
        name=None,  # type: Optional[str]
        **meth_options  # type: Any
    ):  # type: (...) -> None
        """# noqa: D205,D212,D415
        Args:
            constr_data_names: The names of the constraints to aggregate.
                It shall be the output data of other disciplines.
            method_name: The name of the aggregation method, among KS, IKS.
            name: The name of the discipline.
            **meth_options: The options for the aggregation method.

        Raises:
            ValueError: If the method is not supported.
        """
        if method_name not in METHODS_MAP:
            raise ValueError(
                "Unsupported aggregation method named {}".format(method_name)
            )

        super(ConstrAggegationDisc, self).__init__(name)

        self.__method_name = method_name
        self.__input_names = constr_data_names
        self.__meth_options = meth_options
        self.__output_names = [
            "{}_{}".format(self.__method_name, c) for c in constr_data_names
        ]
        self.__out_sizes = {k: 1 for k in self.__output_names}

        self.input_grammar.initialize_from_data_names(self.__input_names)
        self.output_grammar.initialize_from_data_names(self.__output_names)

    def _run(self):  # type: (...) -> None
        c_data = DataConversion.dict_to_array(self.local_data, self.__input_names)
        method = METHODS_MAP[self.__method_name]
        c_agg = atleast_1d(method(c_data, **self.__meth_options))

        out_data = DataConversion.array_to_dict(
            c_agg, self.__output_names, self.__out_sizes
        )
        self.store_local_data(**out_data)

    def _compute_jacobian(
        self,
        inputs=None,  # type: Optional[Sequence[str]]
        outputs=None,  # type: Optional[Sequence[str]]
    ):  # type: (...) -> None
        c_data = DataConversion.dict_to_array(self.local_data, self.__input_names)
        method_jac = METHODS_JAC_MAP[self.__method_name]
        c_agg_jac = method_jac(c_data, **self.__meth_options)

        data_sizes = {k: s.size for k, s in self.local_data.items()}
        data_sizes.update(self.__out_sizes)
        self.jac = DataConversion.jac_2dmat_to_dict(
            c_agg_jac, self.__output_names, self.__input_names, data_sizes
        )
